"""
This file contains the implementations of undoable actions for FlowView.
"""

# prevent circular imports
from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .FlowView import FlowView

from typing import Tuple

from qtpy.QtCore import QObject, QPointF
from qtpy.QtWidgets import QUndoCommand

from ryvencore import InfoMsgs, Flow


def undo_text_multi(items:list, command: str, to_str=None):
    """Generates a text for an undo command that has zero, one or multiple items"""
    
    if to_str is None:
        to_str = str
    
    if len(items) == 1:
        return f'{command} {to_str(items[0]) if to_str else items[0]}'
    elif len(items) == 0:
        return f'Clear-{command}'
    else:
        return f'Multi-{command}'


class FlowUndoCommand(QObject, QUndoCommand):
    """
    The main difference to normal QUndoCommands is the activate feature. This allows the flow widget to add the
    undo command to the undo stack before redo() is called. This is important since some of these commands can cause
    other commands to be added while they are performing redo(), so to prevent those commands to be added to the
    undo stack before the parent command, it is here blocked at first.
    """

    # prevent recursive calls of redo() and undo()
    _any_cmd_active = False

    def __init__(self, flow_view: FlowView) -> None:
        self.flow_view: FlowView = flow_view
        self.flow: Flow = flow_view.flow
        self._activated: bool = False

        QObject.__init__(self)
        QUndoCommand.__init__(self)

    def activate(self):
        self._activated = True
        self.redo()

    def redo(self) -> None:
        if not self._activated:
            return
        elif FlowUndoCommand._any_cmd_active:
            InfoMsgs.write_err(
                'FlowUndoCommand.redo() called while another FlowUndoCommand is active, '
                'most likely due to a recursive redo. This is not allowed. '
                'The editor is now in an undefined state. Save your work and restart the editor. '
            )
            return
        else:
            FlowUndoCommand._any_cmd_active = True
            self.redo_()
            FlowUndoCommand._any_cmd_active = False

    def undo(self) -> None:
        if FlowUndoCommand._any_cmd_active:
            InfoMsgs.write_err(
                'FlowUndoCommand.undo() called while another FlowUndoCommand is active, '
                'most likely due to a recursive undo. This is not allowed. '
                'The editor is now in an undefined state. Save your work and restart the editor. '
            )
            return
        FlowUndoCommand._any_cmd_active = True
        self.undo_()
        FlowUndoCommand._any_cmd_active = False

    def redo_(self):
        """subclassed"""
        pass

    def undo_(self):
        """subclassed"""
        pass


# (not currently used)
# class Nested_Command(FlowUndoCommand):
#     """Simple undo command nesting."""
#     def __init__(self, flow_view, *args):
#         super().__init__(flow_view)
#         self.commands = [arg for arg in args]
#         self.setText('Nested Command')
# 
#     def undo_(self):
#         for command in self.commands:
#             command.undo_()
#   
#     def redo_(self):
#         for command in self.commands:
#             command.redo_()


class Delegate_Command(FlowUndoCommand):
    """
    Custom undo command that delegates the undo and redo actions to two functions.
    """
    def __init__(self, flow_view, text: str, on_undo, on_redo):
        super().__init__(flow_view)
        self.setText(text)
        self._on_undo = on_undo
        self._on_redo = on_redo
    
    def undo_(self):
        self._on_undo()

    def redo_(self):
        self._on_redo()


class MoveComponents_Command(FlowUndoCommand):
    def __init__(self, flow_view, items_list, p_from, p_to):
        super(MoveComponents_Command, self).__init__(flow_view)

        self.items_list = items_list
        self.p_from = p_from
        self.p_to = p_to
        self.last_item_group_pos = p_to
        self.setText(undo_text_multi(self.items_list, 'Move'))

    def undo_(self):
        items_group = self.items_group()
        items_group.setPos(self.p_from)
        self.last_item_group_pos = items_group.pos()
        self.destroy_items_group(items_group)
        for item in self.items_list:
            item.on_move()

    def redo_(self):
        items_group = self.items_group()
        items_group.setPos(self.p_to - self.last_item_group_pos)
        self.destroy_items_group(items_group)
        for item in self.items_list:
            item.on_move()
        
    def items_group(self):
        return self.flow_view.scene().createItemGroup(self.items_list)

    def destroy_items_group(self, items_group):
        self.flow_view.scene().destroyItemGroup(items_group)


class PlaceNode_Command(FlowUndoCommand):
    def __init__(self, flow_view, node_class, pos):
        super().__init__(flow_view)

        self.node_class = node_class
        self.node = None
        self.item_pos = pos
        self._prev_selected = flow_view._current_selected

    def undo_(self):
        self.flow.remove_node(self.node)
        self.flow_view.select_items(self._prev_selected)

    def redo_(self):
        if self.node:
            self.flow.add_node(self.node)
        else:
            self.node = self.flow.create_node(self.node_class)
        self.setText(f'Create {self.node.gui.item}')


class PlaceDrawing_Command(FlowUndoCommand):
    def __init__(self, flow_view, posF, drawing):
        super().__init__(flow_view)

        self.drawing = drawing
        self.drawing_obj_place_pos = posF
        self.drawing_obj_pos = self.drawing_obj_place_pos
        self._prev_selected = flow_view._current_selected

    def undo_(self):
        # The drawing_obj_pos is not anymore the drawing_obj_place_pos because after the
        # drawing object was completed, its actual position got recalculated according to all points and differs from
        # the initial pen press pos (=drawing_obj_place_pos). See DrawingObject.finished().

        self.drawing_obj_pos = self.drawing.pos()

        self.flow_view.remove_component(self.drawing)
        self.flow_view.select_items(self._prev_selected)

    def redo_(self):
        self.flow_view.add_drawing(self.drawing, self.drawing_obj_pos)
        self.setText(self.drawing(True))


class SelectComponents_Command(FlowUndoCommand):
    def __init__(self, flow_view, new_items, prev_items):
        super().__init__(flow_view)
        self.items = new_items
        self.prev_items = prev_items
        self.setText(undo_text_multi(self.items, 'Select'))
        
    def redo_(self):
        self.flow_view.select_items(self.items)
    
    def undo_(self):
        self.flow_view.select_items(self.prev_items)


class RemoveComponents_Command(FlowUndoCommand):
    def __init__(self, flow_view, items):
        super().__init__(flow_view)
        self.items = items
        self.selection = flow_view._current_selected
        self.setText(undo_text_multi(self.items, 'Delete'))
        
        self.broken_connections = (
            set()
        )  # the connections that go beyond the removed nodes and need to be restored in undo
        self.internal_connections = set()

        self.node_items = []
        self.nodes = []
        self.drawings = []

        # importing here to prevent circular imports
        from .nodes.NodeItem import NodeItem
        from .connections.ConnectionItem import ConnectionItem
        from .drawings.DrawingObject import DrawingObject
        
        for i in self.items:
            if isinstance(i, NodeItem):
                self.node_items.append(i)
                self.nodes.append(i.node)
            elif isinstance(i, DrawingObject):
                self.drawings.append(i)

        # for connections
        for i in self.items:
            if not isinstance(i, ConnectionItem):
                continue
            out_port, _ = i.connection
            if out_port.node not in self.nodes:
                self.broken_connections.add(i.connection)

        for n in self.nodes:
            for i in n.inputs:
                cp = n.flow.connected_output(i)
                if cp is not None:
                    cn = cp.node
                    if cn not in self.nodes:
                        self.broken_connections.add((cp, i))
                    else:
                        self.internal_connections.add((cp, i))
            for o in n.outputs:
                for cp in n.flow.connected_inputs(o):
                    cn = cp.node
                    if cn not in self.nodes:
                        self.broken_connections.add((o, cp))
                    else:
                        self.internal_connections.add((o, cp))

    def undo_(self):
        # add nodes
        for n in self.nodes:
            self.flow.add_node(n)

        # add drawings
        for d in self.drawings:
            self.flow_view.add_drawing(d)

        # add connections
        self.restore_broken_connections()
        self.restore_internal_connections()
        
        # removing the items means something was probably selected
        # restore the selection
        self.flow_view.select_items(self.selection)

    def redo_(self):
        # remove connections
        self.remove_broken_connections()
        self.remove_internal_connections()

        # remove nodes
        for n in self.nodes:
            self.flow.remove_node(n)

        # remove drawings
        for d in self.drawings:
            self.flow_view.remove_drawing(d)

    def restore_internal_connections(self):
        for c in self.internal_connections:
            self.flow.add_connection(c)

    def remove_internal_connections(self):
        for c in self.internal_connections:
            self.flow.remove_connection(c)

    def restore_broken_connections(self):
        for c in self.broken_connections:
            self.flow.add_connection(c)

    def remove_broken_connections(self):
        for c in self.broken_connections:
            self.flow.remove_connection(c)


class ConnectPorts_Command(FlowUndoCommand):
    def __init__(self, flow_view, out, inp):
        super().__init__(flow_view)

        # CAN ALSO LEAD TO DISCONNECT INSTEAD OF CONNECT!!

        self.out = out
        self.inp = inp
        self.connection = None
        self.connecting = True

        for i in flow_view.flow.connected_inputs(out):
            if i == self.inp:
                self.connection = (out, i)
                self.connecting = False

    def undo_(self):
        if self.connecting:
            # remove connection
            self.flow.remove_connection(self.connection)
        else:
            # recreate former connection
            self.flow.add_connection(self.connection)

    def redo_(self):
        if self.connecting:
            if self.connection:
                self.flow.add_connection(self.connection)
            else:
                # connection hasn't been created yet
                self.connection = self.flow.connect_nodes(self.out, self.inp)
            self.setText(f'Connect {self.flow_view.connection_items[self.connection]}')
            
        else:
            # remove existing connection
            self.setText(f'Disconnect {self.flow_view.connection_items[self.connection]}')
            self.flow.remove_connection(self.connection)
        

class Paste_Command(FlowUndoCommand):
    def __init__(self, flow_view, data, offset_for_middle_pos):
        super().__init__(flow_view)

        self.data = data
        self.modify_data_positions(offset_for_middle_pos)
        self.pasted_components = None
        self.setText('Paste')

    def modify_data_positions(self, offset):
        """adds the offset to the components' positions in data"""

        for node in self.data['nodes']:
            node['pos x'] = node['pos x'] + offset.x()
            node['pos y'] = node['pos y'] + offset.y()
        for drawing in self.data['drawings']:
            drawing['pos x'] = drawing['pos x'] + offset.x()
            drawing['pos y'] = drawing['pos y'] + offset.y()

    def redo_(self):
        if self.pasted_components is None:
            self.pasted_components = {}

            # create components
            self.create_drawings()

            (
                self.pasted_components['nodes'],
                self.pasted_components['connections'],
            ) = self.flow.load_components(
                nodes_data=self.data['nodes'],
                conns_data=self.data['connections'],
                output_data=self.data['output data'],
            )

            self.select_new_components_in_view()
        else:
            self.add_existing_components()

    def undo_(self):
        # remove components and their items from flow
        for c in self.pasted_components['connections']:
            self.flow.remove_connection(c)
        for n in self.pasted_components['nodes']:
            self.flow.remove_node(n)
        for d in self.pasted_components['drawings']:
            self.flow_view.remove_drawing(d)

    def add_existing_components(self):
        # add existing components and items to flow
        for n in self.pasted_components['nodes']:
            self.flow.add_node(n)
        for c in self.pasted_components['connections']:
            self.flow.add_connection(c)
        for d in self.pasted_components['drawings']:
            self.flow_view.add_drawing(d)

        self.select_new_components_in_view()

    def select_new_components_in_view(self):
        self.flow_view.clear_selection()
        for d in self.pasted_components['drawings']:
            d.setSelected(True)
        for n in self.pasted_components['nodes']:
            node_item = self.flow_view.node_items[n]
            node_item.setSelected(True)

    def create_drawings(self):
        drawings = []
        for d in self.data['drawings']:
            new_drawing = self.flow_view.create_drawing(d)
            self.flow_view.add_drawing(
                new_drawing, posF=QPointF(d['pos x'], d['pos y'])
            )
            drawings.append(new_drawing)
        self.pasted_components['drawings'] = drawings



from typing import Optional, Tuple, List

from qtpy.QtCore import Qt, QPointF, QPoint, QRectF, QMargins, QMarginsF
from qtpy.QtGui import QColor, QPainter, QBrush, QRadialGradient, QLinearGradient, QPen, QPainterPath, QFont, QPolygon
from qtpy.QtWidgets import QStyle, QStyleOption

from ..utils import pythagoras


#
#   notice: available themes are hardcoded in Ryven for CLI; make sure to update those
#   in case of changes affecting it
#


class FlowTheme:

    name = ''
    type_ = 'dark'

    node_selection_stylesheet__base = '''
QScrollArea {
    border: 0px solid grey;
    border-radius: 0px;
}
NodeWidget {

}
'''
    node_selection_stylesheet = ''

    header_padding = (10, 2, 10, 2)  # (left, top, right, botton)

    exec_conn_color = QColor('#ffffff')
    exec_conn_width = 1.5
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor('#ffffff')
    data_conn_width = 1.5
    data_conn_pen_style = Qt.DashLine

    flow_background_brush = QBrush(QColor('#333333'))
    flow_background_grid: Optional[Tuple[str, QColor, int, int, int]] = None
    flow_highlight_pen_color = QColor('#245d75')

    node_item_shadow_color = QColor('#2b2b2b')

    EXPORT: List[str] = []

    def __init__(self):
        pass

    def load(self, data: dict):
        if data and self.name in data.keys():
            imported = {}
            for k, v in data[self.name].items():
                if v != 'default':
                    imported[k] = v
            self._load(imported)

    def _load(self, imported: dict):
        for k, v in imported.items():

            if k == 'exec connection color':
                self.exec_conn_color = self.hex_to_col(v)
            elif k == 'exec connection width':
                self.exec_conn_width = v
            elif k == 'exec connection pen style':
                self.exec_conn_pen_style = self._parse_pen_style(v)

            elif k == 'data connection color':
                self.data_conn_color = self.hex_to_col(v)
            elif k == 'data connection width':
                self.data_conn_width = v
            elif k == 'data connection pen style':
                self.data_conn_pen_style = self._parse_pen_style(v)

            elif k == 'flow background color':
                self.flow_background_brush.setColor(self.hex_to_col(v))

    def build_node_selection_stylesheet(self):
        return self.node_selection_stylesheet__base + '\n' + self.node_selection_stylesheet

    def paint_NI_title_label(self, node_gui, selected: bool, hovering: bool, painter: QPainter, option: QStyleOption,
                             node_style: str, node_title: str, node_color: QColor, node_item_bounding_rect):
        pass

    def paint_PI_label(self, node_gui, painter: QPainter, option: QStyleOption, type_: str, connected: bool, label_str: str,
                       node_color: QColor, bounding_rect: QRectF):
        pass

    def paint_PI(self, node_gui, painter: QPainter, option: QStyleOption, node_color: QColor, type_: str, connected: bool,
                 rect: QRectF):  # padding, w, h):
        pass

    def paint_NI(self, node_gui,
                 selected: bool, hovered: bool, node_style: str,
                 painter: QPainter, option: QStyleOption,
                 color: QColor, w, h, bounding_rect, title_rect):

        painter.setRenderHint(QPainter.Antialiasing)

        if node_style == 'normal':
            self.draw_NI_normal(node_gui, selected, hovered, painter, color, w, h, bounding_rect, title_rect)
        elif node_style == 'small':
            self.draw_NI_small(node_gui, selected, hovered, painter, color, w, h, bounding_rect)

    def draw_NI_normal(self, node_gui, selected: bool, hovered: bool,
                       painter: QPainter, c: QColor, w, h, bounding_rect, title_rect):
        pass

    def draw_NI_small(self, node_gui, selected: bool, hovered: bool,
                      painter: QPainter, c: QColor, w, h, bounding_rect, background_color=None):
        pass

    def paint_NI_selection_border(self, ni, painter: QPainter, color: QColor, w, h, bounding_rect):
        pen = QPen(self.flow_highlight_pen_color)
        pen.setWidth(3)
        painter.setPen(pen)
        painter.setBrush(Qt.NoBrush)

        size_factor = 1.2

        rect = QRectF(bounding_rect)
        rect.setWidth(w*size_factor)
        rect.setHeight(h*size_factor)
        rect.setX(rect.x() - w*size_factor/2)
        rect.setY(rect.y() - h*size_factor/2)

        painter.drawRoundedRect(rect, 10, 10)

    @staticmethod
    def paint_NI_title_label_default(painter: QPainter, node_style: str, title: str, color: QColor, pen_w: float,
                                     font: QFont, node_item_bounding_rect):
        pen = QPen(color)
        pen.setWidth(pen_w)  # type: ignore

        painter.setPen(pen)
        painter.setFont(font)

        text_rect = node_item_bounding_rect
        text_rect.setTop(text_rect.top())

        if node_style == 'normal':
            painter.drawText(text_rect, Qt.AlignLeft | Qt.AlignVCenter, title)
        elif node_style == 'small':
            painter.drawText(text_rect, Qt.AlignTop | Qt.AlignHCenter, title)

    @staticmethod
    def paint_PI_label_default(painter: QPainter, label_str: str, color: QColor, font: QFont, bounding_rect: QRectF):
        painter.setBrush(Qt.NoBrush)
        pen = QPen(color)
        painter.setPen(pen)
        painter.setFont(font)
        painter.drawText(bounding_rect, Qt.AlignCenter, label_str)

    @staticmethod
    def get_header_rect(node_width, node_height, title_rect):
        header_height = 1.0 * title_rect.height()  # 35 * (self.parent_node.title.count('\n')+1)

        header_rect = QRectF()
        header_rect.setTopLeft(QPointF(-node_width / 2, -node_height / 2))
        header_rect.setWidth(node_width)
        header_rect.setHeight(header_height)
        return header_rect

    @staticmethod
    def interpolate_color(c1, c2, val):
        r1 = c1.red()
        g1 = c1.green()
        b1 = c2.blue()
        a1 = c1.alpha()

        r2 = c2.red()
        g2 = c2.green()
        b2 = c2.blue()
        a2 = c2.alpha()

        r = (r2 - r1) * val + r1
        g = (g2 - g1) * val + g1
        b = (b2 - b1) * val + b1
        a = (a2 - a1) * val + a1

        return QColor(r, g, b, a)

    @staticmethod
    def hex_to_col(hex_str: str) -> QColor:
        """Converts a hex value in format '#xxxxxx[xx]' to QColor using alpha value if [xx] is used."""

        h = hex_str.lstrip('#')

        if len(h) == 6:
            r, g, b = tuple(
                int(h[i:i + 2], 16) for i in (0, 2, 4)
            )
            return QColor(r, g, b)
        elif len(h) == 8:
            r, g, b, a = tuple(
                int(h[i:i + 2], 16) for i in (0, 2, 4, 6)
            )
            return QColor(r, g, b, a)

        raise ValueError(f'Invalid hex color string: {hex_str}')

    @staticmethod
    def col(c: QColor, alpha=255):
        return QColor(c.red(), c.green(), c.blue(), alpha)

    @staticmethod
    def _parse_pen_style(s: str):
        if s == 'solid line':
            return Qt.SolidLine
        elif s == 'dash line':
            return Qt.DashLine
        elif s == 'dash dot line':
            return Qt.DashDotLine
        elif s == 'dash dot dot line':
            return Qt.DashDotDotLine
        elif s == 'dot line':
            return Qt.DotLine


class FlowTheme_Toy(FlowTheme):
    name = 'Toy'

    node_selection_stylesheet = ''

    header_padding = (12, 5, 10, 2)

    exec_conn_color = QColor(188, 187, 242)
    exec_conn_width = 5
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor(188, 187, 242)
    data_conn_width = 5
    data_conn_pen_style = Qt.DashLine

    flow_background_brush = QBrush(QColor('#333333'))

    def paint_NI_title_label(self, node_gui, selected, hovering, painter, option, node_style, node_title, node_color,
                             node_item_bounding_rect):

        if node_style == 'normal':
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=QColor(30, 43, 48) if not hovering else node_color.lighter(),
                pen_w=2 if hovering else 1.5,
                font=QFont('Poppins', 15),
                node_item_bounding_rect=node_item_bounding_rect
            )
        else:
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=QColor(30, 43, 48) if not hovering else node_color.lighter(),
                pen_w=1.5,
                font=QFont('K2D', 20, QFont.Bold, True),
                node_item_bounding_rect=node_item_bounding_rect
            )

    def paint_PI_label(self, node_gui, painter, option, type_, connected, label_str, node_color, bounding_rect):
        c = QColor('#FFFFFF')
        self.paint_PI_label_default(painter, label_str, c, QFont("Source Code Pro", 10, QFont.Bold), bounding_rect)

    def paint_PI(self, node_gui, painter, option, node_color, type_, connected, rect):

        color = QColor('#2E688C') if type_ == 'data' else QColor('#3880ad')
        if option.state & QStyle.State_MouseOver:
            color = color.lighter()

        brush = QBrush(QColor(color))
        painter.setBrush(brush)
        painter.setPen(Qt.NoPen)

        painter.drawEllipse(rect)

    def draw_NI_normal(self, node_gui, selected: bool, hovered: bool,
                       painter: QPainter, c: QColor, w, h, bounding_rect, title_rect):

        # main rect
        header_color = QColor(
            int(c.red() / 10 + 100),
            int(c.green() / 10 + 100),
            int(c.blue() / 10 + 100),
        )
        if selected:
            header_color = header_color.lighter()
        body_gradient = QRadialGradient(bounding_rect.topLeft(), pythagoras(h, w))
        body_gradient.setColorAt(0, self.col(header_color, alpha=200))
        body_gradient.setColorAt(1, self.col(header_color, alpha=0))

        painter.setBrush(body_gradient)
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, 12, 12)

        header_gradient = QLinearGradient(FlowTheme_Toy.get_header_rect(w, h, title_rect).topRight(),
                                          FlowTheme_Toy.get_header_rect(w, h, title_rect).bottomLeft())
        header_gradient.setColorAt(0, QColor(c.red(), c.green(), c.blue(), 255))
        header_gradient.setColorAt(1, QColor(c.red(), c.green(), c.blue(), 0))
        painter.setBrush(header_gradient)
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(FlowTheme_Toy.get_header_rect(w, h, title_rect), 12, 12)

    def draw_NI_small(self, node_gui, selected: bool, hovered: bool,
                      painter: QPainter, c: QColor, w, h, bounding_rect, background_color=None):

        path = QPainterPath()
        path.moveTo(-w / 2, 0)

        path.cubicTo(-w / 2, -h / 2,
                     -w / 2, -h / 2,
                     0, -h / 2)
        path.cubicTo(+w / 2, -h / 2,
                     +w / 2, -h / 2,
                     +w / 2, 0)
        path.cubicTo(+w / 2, +h / 2,
                     +w / 2, +h / 2,
                     0, +h / 2)
        path.cubicTo(-w / 2, +h / 2,
                     -w / 2, +h / 2,
                     -w / 2, 0)
        path.closeSubpath()

        body_gradient = QLinearGradient(bounding_rect.bottomLeft(),
                                        bounding_rect.topRight())
        body_gradient.setColorAt(0, QColor(c.red(), c.green(), c.blue(), 150))
        body_gradient.setColorAt(1, QColor(c.red(), c.green(), c.blue(), 80))

        painter.setBrush(body_gradient)
        painter.setPen(QPen(QColor(30, 43, 48)))

        painter.drawPath(path)


class FlowTheme_DarkTron(FlowTheme):
    name = 'Tron'

    node_selection_stylesheet = ''

    header_padding = (12, 5, 10, 2)

    exec_conn_color = QColor(0, 120, 180)
    exec_conn_width = 4
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor(0, 120, 180)
    data_conn_width = 4
    data_conn_pen_style = Qt.DashLine

    flow_background_brush = QBrush(QColor('#333333'))

    def paint_NI_title_label(self, node_gui, selected, hovering, painter, option, node_style, node_title, node_color,
                             node_item_bounding_rect):
        if node_style == 'normal':
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=node_color if not (hovering or selected) else node_color.lighter().lighter(),
                pen_w=2,
                font=QFont('Poppins', 15),
                node_item_bounding_rect=node_item_bounding_rect
            )
        else:
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=node_color,
                pen_w=2,
                font=QFont('K2D', 20, QFont.Bold, True),
                node_item_bounding_rect=node_item_bounding_rect
            )

    def paint_PI_label(self, node_gui, painter, option, type_, connected, label_str, node_color, bounding_rect):
        if type_ == 'exec':
            c = QColor('#FFFFFF')
        else:
            c = node_color

        self.paint_PI_label_default(painter, label_str, c, QFont("Source Code Pro", 10, QFont.Bold), bounding_rect)

    def paint_PI(self, node_gui, painter, option, node_color, type_, connected, rect):

        color = QColor('#FFFFFF') if type_ == 'exec' else QColor(node_color)
        pen = QPen(color)
        pen.setWidth(2)
        painter.setPen(pen)
        if connected or \
                option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
            r = color.red()
            g = color.green()
            b = color.blue()
            brush = QBrush(QColor(r, g, b, 100))
            painter.setBrush(brush)
        else:
            painter.setBrush(Qt.NoBrush)

        painter.drawEllipse(rect)

    # def paint_NI(self, node, node_style,
    #              painter, option,
    #              color: QColor, w: int, h: int, bounding_rect, title_rect):
    #
    #     painter.setRenderHint(QPainter.Antialiasing)
    #
    #     if node_style == 'normal':
    #         self.draw_NI_normal(node, painter, color, w, h, bounding_rect, title_rect)
    #     elif node_style == 'small':
    #         if option.state & QStyle.State_MouseOver:  # use special dark background color when mouse hovers
    #             self.draw_NI_small(node, painter, color, w, h, bounding_rect, color.darker())
    #         else:
    #             self.draw_NI_small(node, painter, color, w, h, bounding_rect, QColor('#212429'))

    def draw_NI_normal(self, node_gui, selected: bool, hovered: bool,
                       painter, c: QColor, w: int, h: int, bounding_rect, title_rect):

        background_color = QColor('#212224')
        painter.setBrush(background_color)
        pen = QPen(c if not selected else c.lighter())
        pen.setWidth(2)
        painter.setPen(pen)
        body_path = self.get_extended_body_path(w, h)
        painter.drawPath(body_path)

        header_gradient = QLinearGradient(self.get_header_rect(w, h, title_rect).topRight(),
                                          self.get_header_rect(w, h, title_rect).bottomLeft())
        header_gradient.setColorAt(0, QColor(c.red(), c.green(), c.blue(), 255))
        header_gradient.setColorAt(0.5, QColor(c.red(), c.green(), c.blue(), 100))
        header_gradient.setColorAt(1, QColor(c.red(), c.green(), c.blue(), 0))
        painter.setBrush(header_gradient)
        header_path = self.get_extended_header_path(w, h, title_rect)
        painter.drawPath(header_path)

    @staticmethod
    def get_extended_body_path(w, h):

        c_s = 10  # corner size

        path = QPainterPath()
        path.moveTo(+w / 2, -h / 2 + c_s)
        path.lineTo(+w / 2 - c_s, -h / 2)
        path.lineTo(-w / 2 + c_s, -h / 2)
        path.lineTo(-w / 2, -h / 2 + c_s)
        path.lineTo(-w / 2, +h / 2 - c_s)
        path.lineTo(-w / 2 + c_s, +h / 2)
        path.lineTo(+w / 2 - c_s, +h / 2)
        path.lineTo(+w / 2, +h / 2 - c_s)
        path.closeSubpath()
        return path

    def get_extended_header_path(self, w, h, title_rect):

        c_s = 10  # corner size

        # header_height = 35 * (NIPainter_DarkTron.ni.parent_node.title.count('\n') + 1)
        header_height = self.get_header_rect(w, h, title_rect).height()
        header_bottom = -h / 2 + header_height
        path = QPainterPath()
        path.moveTo(+w / 2, -h / 2 + c_s)
        path.lineTo(+w / 2 - c_s, -h / 2)
        path.lineTo(-w / 2 + c_s, -h / 2)
        path.lineTo(-w / 2, -h / 2 + c_s)
        path.lineTo(-w / 2, header_bottom - c_s)
        path.lineTo(-w / 2 + c_s, header_bottom)
        path.lineTo(+w / 2 - c_s, header_bottom)
        path.lineTo(+w / 2, header_bottom - c_s)
        path.closeSubpath()
        return path

    def draw_NI_small(self, node_gui, selected: bool, hovered: bool,
                      painter: QPainter, c: QColor, w, h, bounding_rect, background_color=None):

        if hovered:
            background_color = c.darker()
        else:
            background_color = QColor('#212429')

        c_s = 10  # corner size

        path = QPainterPath()
        path.moveTo(-w / 2, 0)

        path.lineTo(-w / 2 + c_s / 2, -h / 2 + c_s / 2)
        path.lineTo(0, -h / 2)
        path.lineTo(+w / 2 - c_s / 2, -h / 2 + c_s / 2)
        path.lineTo(+w / 2, 0)
        path.lineTo(+w / 2 - c_s / 2, +h / 2 - c_s / 2)
        path.lineTo(0, +h / 2)
        path.lineTo(-w / 2 + c_s / 2, +h / 2 - c_s / 2)
        path.closeSubpath()

        painter.setBrush(background_color)
        pen = QPen(c)
        pen.setWidth(2)
        painter.setPen(pen)

        painter.drawPath(path)


class FlowTheme_Ghost(FlowTheme):
    name = 'Ghost'
    type_ = 'dark'

    node_selection_stylesheet = ''

    header_padding = (12, 6, 10, 2)

    exec_conn_color = QColor(0, 17, 25)
    exec_conn_width = 2
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor(0, 17, 25)
    data_conn_width = 2
    data_conn_pen_style = Qt.DashLine

    flow_background_color = QColor('#333333')
    flow_background_brush = QBrush(flow_background_color)

    node_color = QColor(28, 28, 28, 170)
    node_small_color = QColor('#212429')

    EXPORT = [
        'nodes color',
        'small nodes color',
        'flow background color'
    ]

    def _load(self, imported: dict):
        super()._load(imported)

        for k, v in imported.items():
            if k == 'nodes color':
                self.node_color = self.hex_to_col(v)
            elif k == 'small nodes color':
                self.node_small_color = self.hex_to_col(v)
            elif k == 'flow background color':
                self.flow_background_color = self.hex_to_col(v)
                self.flow_background_brush = QBrush(self.flow_background_color)

    def paint_NI_title_label(self, node_gui, selected, hovering, painter, option, node_style, node_title, node_color,
                             node_item_bounding_rect):

        if node_style == 'normal':
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=node_color if not hovering else node_color.lighter(),
                pen_w=2,
                font=QFont('Poppins', 15),
                node_item_bounding_rect=node_item_bounding_rect
            )
        else:
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=node_color,
                pen_w=2,
                font=QFont('K2D', 20, QFont.Bold, True),
                node_item_bounding_rect=node_item_bounding_rect
            )

    def paint_PI_label(self, node_gui, painter, option, type_, connected, label_str, node_color, bounding_rect):
        if type_ == 'exec':
            c = QColor('#FFFFFF')
        else:
            c = node_color

        self.paint_PI_label_default(painter, label_str, c, QFont("Source Code Pro", 10, QFont.Bold), bounding_rect)

    def paint_PI(self, node_gui, painter, option, node_color, type_, connected, rect):

        color = QColor('#FFFFFF') if type_ == 'exec' else QColor(node_color)

        if type_ == 'exec':
            if connected or \
                    option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
                brush = QBrush(QColor(255, 255, 255, 100))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)
        elif type_ == 'data':
            if connected or \
                    option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
                r = color.red()
                g = color.green()
                b = color.blue()
                brush = QBrush(QColor(r, g, b, 100))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)

        pen = QPen(color)
        pen.setWidth(1)
        painter.setPen(pen)
        painter.drawEllipse(rect.marginsRemoved(QMarginsF(3, 3, 3, 3)))

    # def paint_NI(self, node, node_style,
    #              painter, option,
    #              color: QColor, w: int, h: int, bounding_rect, title_rect):
    #
    #     painter.setRenderHint(QPainter.Antialiasing)
    #
    #     if node_style == 'normal':
    #         self.draw_NI_normal(node, painter, color, w, h, bounding_rect, title_rect)
    #     elif node_style == 'small':
    #         if option.state & QStyle.State_MouseOver:  # use special dark background color when mouse hovers
    #             self.draw_NI_small(node, painter, color, w, h, bounding_rect, background_color=color.darker())
    #         else:
    #             self.draw_NI_small(node, painter, color, w, h, bounding_rect)

    def draw_NI_normal(self, node_gui, selected: bool, hovered: bool,
                       painter, c, w, h, bounding_rect, title_rect):

        background_color = self.node_color

        # body_gradient = QRadialGradient(QPointF(bounding_rect.topLeft().x() + w,
        #                                         bounding_rect.topLeft().y() - h),
        #                                 pythagoras(2 * h, 2 * w))
        # body_gradient.setColorAt(0, QColor(c.red() / 10 + 100, c.green() / 10 + 100, c.blue() / 10 + 100, 100))
        # body_gradient.setColorAt(0.7, background_color)
        # body_gradient.setColorAt(1, background_color)
        #
        # painter.setBrush(body_gradient)

        painter.setBrush(background_color)
        pen = QPen(c.darker())
        pen.setWidth(1 if not selected else 5)
        painter.setPen(pen)
        body_path = self.get_extended_body_path(5, w, h)
        painter.drawPath(body_path)

    @staticmethod
    def get_extended_body_path(c_s, w, h):

        path = QPainterPath()
        path.moveTo(+w / 2, -h / 2 + c_s)
        path.lineTo(+w / 2 - c_s, -h / 2)
        path.lineTo(-w / 2 + c_s, -h / 2)
        path.lineTo(-w / 2, -h / 2 + c_s)
        path.lineTo(-w / 2, +h / 2 - c_s)
        path.lineTo(-w / 2 + c_s, +h / 2)
        path.lineTo(+w / 2 - c_s, +h / 2)
        path.lineTo(+w / 2, +h / 2 - c_s)
        path.closeSubpath()
        return path

    def draw_NI_small(self, node_gui, selected: bool, hovered: bool,
                      painter, c, w, h, bounding_rect, background_color=None):

        background_color = self.node_small_color
        c_s = 10  # corner size

        path = self.get_extended_body_path(c_s, w, h)  # equals the small in this case

        painter.setBrush(background_color)
        # pen = QPen(QColor('#333333'))  # QPen(c)
        # pen.setWidth(1)
        # painter.setPen(pen)
        painter.setPen(Qt.NoPen)

        painter.drawPath(path)


class FlowTheme_Blender(FlowTheme):
    name = 'Blender'

    node_selection_stylesheet = ''

    header_padding = (5, 0, 0, 0)

    exec_conn_color = QColor(0, 17, 25)
    exec_conn_width = 3
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor(200, 200, 200)
    data_conn_width = 2.5
    data_conn_pen_style = Qt.SolidLine

    flow_background_color = QColor('#232323')
    flow_background_brush = QBrush(flow_background_color)

    node_color = QColor('#3f3f3f')
    corner_radius_normal = 5
    corner_radius_small = 10

    EXPORT = [
        'nodes color',
        'flow background color'
    ]

    def _load(self, imported: dict):
        super()._load(imported)

        for k, v in imported.items():
            if k == 'nodes color':
                self.node_color = self.hex_to_col(v)
            elif k == 'flow background color':
                self.flow_background_color = self.hex_to_col(v)
                self.flow_background_brush = QBrush(self.flow_background_color)

    def paint_NI_title_label(self, node_gui, selected, hovering, painter, option, node_style, node_title, node_color,
                             node_item_bounding_rect):
        if node_style == 'normal':
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=QColor('#FFFFFF'),
                pen_w=2,
                font=QFont('Segoe UI', 11),
                node_item_bounding_rect=node_item_bounding_rect
            )
        else:
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=node_color,
                pen_w=2,
                font=QFont('Segoe UI', 15, QFont.Bold, True),
                node_item_bounding_rect=node_item_bounding_rect
            )

    def paint_PI_label(self, node_gui, painter, option, type_, connected, label_str, node_color, bounding_rect):
        if type_ == 'exec':
            c = QColor('#FFFFFF')
        else:
            c = node_color

        self.paint_PI_label_default(painter, label_str, c, QFont("Source Code Pro", 10, QFont.Bold), bounding_rect)

    def paint_PI(self, node_gui, painter, option, node_color, type_, connected, rect):

        color = QColor('#FFFFFF') if type_ == 'exec' else node_color

        painter.setBrush(QBrush(color))

        # if type_ == 'exec':
        #     if connected or \
        #             option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
        #         brush = QBrush(QColor(255, 255, 255, 100))
        #         painter.setBrush(brush)
        #     else:
        #         painter.setBrush(Qt.NoBrush)
        # elif type_ == 'data':
        #     if connected or \
        #             option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
        #         r = node_color.red()
        #         g = node_color.green()
        #         b = node_color.blue()
        #         brush = QBrush(QColor(r, g, b, 100))
        #         painter.setBrush(brush)
        #     else:
        #         painter.setBrush(Qt.NoBrush)

        pen = QPen(color)
        pen.setWidth(1)
        painter.setPen(pen)
        painter.drawEllipse(rect.marginsRemoved(QMarginsF(2, 2, 2, 2)))

    def draw_NI_normal(self, node_gui, selected: bool, hovered: bool,
                       painter, c, w, h, bounding_rect, title_rect):

        background_color = self.node_color
        header_color = QColor(c.red(), c.green(), c.blue(), 180)
        if selected:
            header_color = header_color.lighter()


        rel_header_height = self.get_header_rect(w, h, title_rect).height() / h
        gradient = QLinearGradient(bounding_rect.topLeft(), bounding_rect.bottomLeft())
        gradient.setColorAt(0, header_color)
        gradient.setColorAt(rel_header_height, header_color)
        gradient.setColorAt(rel_header_height + 0.0001, background_color)
        gradient.setColorAt(1, background_color)

        painter.setBrush(gradient)
        painter.setPen(
            Qt.NoPen
            if not selected else
            QPen(QColor(200, 200, 200))
        )
        painter.drawRoundedRect(bounding_rect, self.corner_radius_normal, self.corner_radius_normal)

    def draw_NI_small(self, node_gui, selected: bool, hovered: bool,
                      painter, c, w, h, bounding_rect, background_color=None):

        background_color = QColor('#212429')
        painter.setBrush(self.interpolate_color(c, background_color, 0.97))
        painter.setPen(
            QPen(c)
            if not selected else
            QPen(QColor(200, 200, 200))
        )
        painter.drawRoundedRect(bounding_rect, self.corner_radius_small, self.corner_radius_small)


class FlowTheme_Simple(FlowTheme):
    name = 'Simple'
    type_ = 'dark'

    node_selection_stylesheet = ''

    header_padding = (10, 2, 10, 2)

    exec_conn_color = QColor('#989c9f')
    exec_conn_width = 2
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor('#989c9f')
    data_conn_width = 2
    data_conn_pen_style = Qt.DashLine

    flow_background_color = QColor('#3f4044')
    flow_background_brush = QBrush(flow_background_color)

    node_background_color = QColor('#212429')
    node_small_background_color = QColor('#212429')

    EXPORT = [
        'nodes background color',
        'small nodes background color',
        'flow background color'
    ]

    def _load(self, imported: dict):
        super()._load(imported)

        for k, v in imported.items():
            if k == 'nodes background color':
                self.node_background_color = self.hex_to_col(v)
            elif k == 'small nodes background color':
                self.node_small_background_color = self.hex_to_col(v)
            elif k == 'flow background color':
                self.flow_background_color = self.hex_to_col(v)
                self.flow_background_brush = QBrush(self.flow_background_color)

    def paint_NI_title_label(self, node_gui, selected, hovering, painter, option, node_style, node_title, node_color,
                             node_item_bounding_rect):
        if node_style == 'normal':
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=QColor('#312b29'),
                pen_w=2,
                font=QFont('ASAP', 13, QFont.Bold),
                node_item_bounding_rect=node_item_bounding_rect
            )
        else:
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=node_color,
                pen_w=2,
                font=QFont('Poppins', 15, QFont.Thin),
                node_item_bounding_rect=node_item_bounding_rect
            )

    def paint_PI_label(self, node_gui, painter, option, type_, connected, label_str, node_color, bounding_rect):
        c = None
        if not connected:
            c = QColor('#53585c')
        else:
            if type_ == 'exec':
                c = QColor('#cccccc')
            else:
                c = node_color

        self.paint_PI_label_default(painter, label_str, c, QFont("Courier New", 10, QFont.Bold), bounding_rect)

    def paint_PI(self, node_gui, painter, option, node_color, type_, connected, rect):

        color = None
        if not connected:
            color = QColor('#53585c')
        else:
            if type_ == 'exec':
                color = QColor('#dddddd')
            else:
                color = QColor(node_color)

        if type_ == 'exec':
            if connected or \
                    option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
                brush = QBrush(QColor(255, 255, 255, 100))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)
        elif type_ == 'data':
            if connected or \
                    option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
                r = color.red()
                g = color.green()
                b = color.blue()
                brush = QBrush(QColor(r, g, b, 100))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)

        brush = QBrush(color)
        painter.setBrush(brush)
        painter.setPen(Qt.NoPen)

        painter.drawEllipse(rect.marginsRemoved(QMarginsF(2, 2, 2, 2)))
        # painter.drawEllipse(QRectF(padding+w/8, padding+h/8, 3*w/4, 3*h/4))

    def draw_NI_normal(self, node_gui, selected: bool, hovered: bool,
                       painter, c, w, h, bounding_rect, title_rect):

        background_color = self.node_background_color

        if selected:
            header_color = c.lighter()
        else:
            header_color = c

        rel_header_height = self.get_header_rect(w, h, title_rect).height() / h
        gradient = QLinearGradient(bounding_rect.topLeft(), bounding_rect.bottomLeft())
        gradient.setColorAt(0, header_color)
        gradient.setColorAt(rel_header_height, header_color)
        gradient.setColorAt(rel_header_height + 0.0001, background_color)
        gradient.setColorAt(1, background_color)

        painter.setBrush(gradient)
        painter.setPen(Qt.NoPen)  # QPen(c.darker()))
        painter.drawRoundedRect(bounding_rect, 9, 9)

    def draw_NI_small(self, node_gui, selected: bool, hovered: bool,
                      painter, c, w, h, bounding_rect, background_color=None):

        background_color = self.node_small_background_color
        c_s = 10
        painter.setBrush(self.interpolate_color(c, background_color, 0.97))
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, c_s, c_s)


class FlowTheme_Ueli(FlowTheme):
    name = 'Ueli'

    node_selection_stylesheet = ''

    header_padding = (5, 2, 10, 0)

    exec_conn_color = QColor('#989c9f')
    exec_conn_width = 2
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor('#989c9f')
    data_conn_width = 2
    data_conn_pen_style = Qt.DashLine

    flow_background_color = QColor('#3f4044')
    flow_background_brush = QBrush(flow_background_color)

    nodes_background_color = QColor('#212429')
    small_nodes_background_color = nodes_background_color

    EXPORT = [
        'nodes background color',
        'small nodes background color',
        'flow background color'
    ]

    def _load(self, imported: dict):
        super()._load(imported)

        for k, v in imported.items():
            if k == 'nodes background color':
                c = self.hex_to_col(v)
                self.nodes_background_color = c
            elif k == 'small nodes background color':
                self.small_nodes_background_color = self.hex_to_col(v)
            elif k == 'flow background color':
                self.flow_background_color = self.hex_to_col(v)
                self.flow_background_brush = QBrush(self.flow_background_color)

    def paint_NI_title_label(self, node_gui, selected, hovering, painter, option, node_style, node_title, node_color,
                             node_item_bounding_rect):
        if node_style == 'normal':
            painter.setPen(QPen(QColor(node_color.name())))
            painter.setFont(QFont('Poppins', 13))
            painter.drawText(node_item_bounding_rect, Qt.AlignLeft | Qt.AlignVCenter, node_title)
        else:
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=node_color,
                pen_w=2,
                font=QFont('Poppins', 15, QFont.Thin),
                node_item_bounding_rect=node_item_bounding_rect
            )

    def paint_PI_label(self, node_gui, painter, option, type_, connected, label_str, node_color, bounding_rect):

        c = None
        if not connected:
            c = '#53585c'
        else:
            if type_ == 'exec':
                c = '#cccccc'
            else:
                c = node_color
        color = QColor(c)

        self.paint_PI_label_default(painter, label_str, color, QFont("Courier New", 10, QFont.Bold), bounding_rect)

    def paint_PI(self, node_gui, painter, option, node_color, type_, connected, rect):

        color_str = None
        if not connected:
            color_str = '#53585c'
        else:
            if type_ == 'exec':
                color_str = '#dddddd'
            else:
                color_str = node_color

        color = QColor(color_str)
        if type_ == 'exec':
            if connected or \
                    option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
                brush = QBrush(QColor(255, 255, 255, 100))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)
        elif type_ == 'data':
            if connected or \
                    option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
                r = color.red()
                g = color.green()
                b = color.blue()
                brush = QBrush(QColor(r, g, b, 100))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)

        brush = QBrush(color)
        painter.setBrush(brush)
        painter.setPen(Qt.NoPen)

        painter.drawEllipse(rect.marginsRemoved(QMarginsF(2, 2, 2, 2)))
        # painter.drawEllipse(QRectF(padding+w/8, padding+h/8, 3*w/4, 3*h/4))

    def draw_NI_normal(self, node_gui, selected: bool, hovered: bool,
                       painter, c, w, h, bounding_rect: QRectF, title_rect):

        if selected:
            background_color = self.interpolate_color(self.nodes_background_color, c.darker(), 0.18)
        else:
            background_color = self.nodes_background_color

        header_color = c

        header_height = self.get_header_rect(w, h, title_rect).height()
        rel_header_height = header_height / h
        gradient = QLinearGradient(bounding_rect.topLeft(), bounding_rect.bottomLeft())
        gradient.setColorAt(0, header_color)
        gradient.setColorAt(rel_header_height, header_color)
        gradient.setColorAt(rel_header_height + 0.0001, background_color)
        gradient.setColorAt(1, background_color)

        painter.setBrush(QBrush(background_color))
        painter.setPen(Qt.NoPen)  # QPen(c.darker()))
        painter.drawRoundedRect(QRectF(
            QPointF(bounding_rect.left(), bounding_rect.top() + header_height),
            bounding_rect.bottomRight()
        ), 6, 6)

    def draw_NI_small(self, node_gui, selected: bool, hovered: bool,
                      painter, c, w, h, bounding_rect, background_color=None):
        background_color = self.small_nodes_background_color
        c_s = 10  # corner size
        painter.setBrush(self.interpolate_color(c, background_color, 0.97))
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, c_s, c_s)


class FlowTheme_PureDark(FlowTheme):
    name = 'pure dark'

    node_selection_stylesheet = ''

    header_padding = (4, 2, 2, 2)

    exec_conn_color = QColor('#ffffff')
    exec_conn_width = 1.5
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor('#ffffff')
    data_conn_width = 1.5
    data_conn_pen_style = Qt.DashLine

    flow_background_brush = QBrush(QColor('#1E242A'))
    flow_background_grid = ('points', flow_background_brush.color().lighter(), 2, 50, 50)

    node_item_shadow_color = QColor('#101010')

    node_normal_bg_col = QColor('#0C1116')
    node_small_bg_col = QColor('#363c41')
    node_title_color = QColor('#ffffff')
    port_pin_pen_color = QColor('#ffffff')

    EXPORT = [
        'extended node background color',
        'small node background color',
        'node title color',
        'port pin pen color'
    ]

    def _load(self, imported: dict):
        super()._load(imported)

        for k, v in imported.items():
            if k == 'extended node background color':
                self.node_normal_bg_col = self.hex_to_col(v)
            elif k == 'small node background color':
                self.node_small_bg_col = self.hex_to_col(v)
            elif k == 'node title color':
                self.node_title_color = self.hex_to_col(v)
            elif k == 'port pin pen color':
                self.port_pin_pen_color = self.hex_to_col(v)

    def paint_NI_title_label(self, node_gui, selected, hovering, painter, option, node_style, node_title, node_color,
                             node_item_bounding_rect):

        painter.setPen(QPen(self.node_title_color))

        if node_style == 'normal':
            painter.setFont(QFont('Segoe UI', 11))
            align = Qt.AlignLeft | Qt.AlignVCenter
        else:
            painter.setFont(QFont('Segoe UI', 15))
            align = Qt.AlignCenter

        painter.drawText(node_item_bounding_rect, align, node_title)

    def paint_PI_label(self, node_gui, painter, option, type_, connected, label_str, node_color, bounding_rect):
        c = None
        if not connected:
            c = QColor('#53585c')
        else:
            if type_ == 'exec':
                c = QColor('#cccccc')
            else:
                c = node_color

        self.paint_PI_label_default(painter, label_str, c, QFont("Segoe UI", 10), bounding_rect)

    def paint_PI(self, node_gui, painter, option, node_color, type_, connected, rect):

        if connected:
            painter.setBrush(QColor('#508AD8'))
            painter.setPen(Qt.NoPen)
        else:
            painter.setBrush(Qt.NoBrush)
            p = QPen(self.port_pin_pen_color)
            p.setWidthF(1.1)
            painter.setPen(p)

        # rect = QRectF(padding + w / 8, padding + h / 8, 3 * w / 4, 3 * h / 4)
        if type_ == 'exec':
            painter.setBrush(QBrush(QColor('white')))
        # painter.drawEllipse(rect)
        painter.drawEllipse(rect.marginsRemoved(QMarginsF(2, 2, 2, 2)))

    def draw_NI_normal(self, node_gui, selected: bool, hovered: bool,
                       painter, c, w, h, bounding_rect: QRectF, title_rect):

        if selected:
            background_color = self.interpolate_color(self.node_normal_bg_col, c.darker(), 0.18)
        else:
            background_color = self.node_normal_bg_col

        header_height = self.get_header_rect(w, h, title_rect).height()

        painter.setBrush(QBrush(background_color))
        painter.setPen(Qt.NoPen)  # QPen(c.darker()))
        painter.drawRoundedRect(QRectF(
            QPointF(bounding_rect.left(), bounding_rect.top() + header_height),
            bounding_rect.bottomRight()
        ), 3, 3)

        p = QPen(c)
        p.setWidthF(2.3)
        painter.setPen(p)
        painter.drawLine(
            QPointF(bounding_rect.left(), bounding_rect.top() + header_height),
            QPointF(bounding_rect.right(), bounding_rect.top() + header_height)
        )

    def draw_NI_small(self, node_gui, selected: bool, hovered: bool,
                      painter, c, w, h, bounding_rect, background_color=None):

        painter.setBrush(QBrush(self.node_small_bg_col))
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, 4, 4)


class FlowTheme_PureLight(FlowTheme_PureDark):
    name = 'pure light'
    type_ = 'light'

    header_padding = (2, 2, 2, 2)

    node_selection_stylesheet = '''
NodeSelectionWidget {
    background-color: white;
}
NodeWidget {
    background-color: white;
}
    '''

    exec_conn_color = QColor('#1f1f1f')

    data_conn_color = QColor('#1f1f1f')

    flow_background_brush = QBrush(QColor('#ffffff'))
    flow_background_grid = ('points', QColor('#dddddd'), 2, 20, 20)

    node_normal_bg_col = QColor('#cdcfd1')
    node_small_bg_col = QColor('#bebfc1')
    node_title_color = QColor('#1f1f1f')
    port_pin_pen_color = QColor('#1f1f1f')

    node_item_shadow_color = QColor('#cccccc')


class FlowTheme_Colorful(FlowTheme):
    name = 'colorful dark'

    header_padding = (12, 0, 2, 2)

    exec_conn_color = QColor('#ffffff')
    exec_conn_width = 1.5
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor('#ffffff')
    data_conn_width = 1.5
    data_conn_pen_style = Qt.DashLine

    flow_background_brush = QBrush(QColor('#1E242A'))
    flow_background_grid = ('points', flow_background_brush.color().lighter(), 2, 50, 50)

    node_ext_background_color = QColor('#0C1116')
    node_small_background_color = QColor('#363c41')
    node_title_color = QColor('#ffffff')
    port_pin_pen_color = QColor('#ffffff')

    EXPORT = [
        'node title color',
        'port pin pen color'
    ]

    def _load(self, imported: dict):
        super()._load(imported)

        for k, v in imported.items():
            if k == 'node title color':
                self.node_title_color = self.hex_to_col(v)
            elif k == 'port pin pen color':
                self.port_pin_pen_color = self.hex_to_col(v)

    def paint_NI_title_label(self, node_gui, selected, hovering, painter, option, node_style, node_title, node_color,
                             node_item_bounding_rect):

        painter.setPen(QPen(self.node_title_color))

        if node_style == 'normal':
            painter.setFont(QFont('Segoe UI', 11))
            align = Qt.AlignLeft | Qt.AlignVCenter
        else:
            painter.setFont(QFont('Segoe UI', 15))
            align = Qt.AlignCenter

        painter.drawText(node_item_bounding_rect, align, node_title)

    def paint_PI_label(self, node_gui, painter, option, type_, connected, label_str, node_color, bounding_rect):
        c = None
        if not connected:
            c = QColor('#dddddd')
        else:
            if type_ == 'exec':
                c = QColor('#cccccc')
            else:
                c = node_color

        self.paint_PI_label_default(painter, label_str, c, QFont("Segoe UI", 10), bounding_rect)

    def paint_PI(self, node_gui, painter, option, node_color, type_, connected, rect):

        if connected:
            painter.setBrush(QColor('#508AD8'))
            painter.setPen(Qt.NoPen)
        else:
            painter.setBrush(Qt.NoBrush)
            p = QPen(self.port_pin_pen_color)
            p.setWidthF(1.1)
            painter.setPen(p)

        painter.drawEllipse(rect.marginsRemoved(QMarginsF(2, 2, 2, 2)))
        # painter.drawEllipse(QRectF(padding + w / 8, padding + h / 8, 3 * w / 4, 3 * h / 4))

    def draw_NI_normal(self, node_gui, selected: bool, hovered: bool,
                       painter, c, w, h, bounding_rect: QRectF, title_rect):

        background_color = c
        background_color.setAlpha(150)

        if selected:
            header_color = QColor(c.red(), c.green(), c.blue(), 130)
        else:
            header_color = QColor(c.red(), c.green(), c.blue(), 130).darker()

        rel_header_height = self.get_header_rect(w, h, title_rect).height() / h
        gradient = QLinearGradient(bounding_rect.topLeft(), bounding_rect.bottomLeft())
        gradient.setColorAt(0, header_color)
        gradient.setColorAt(rel_header_height, header_color)
        gradient.setColorAt(rel_header_height + 0.0001, background_color)
        gradient.setColorAt(1, background_color)

        painter.setBrush(gradient)
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, 7, 7)

    def draw_NI_small(self, node_gui, selected: bool, hovered: bool,
                      painter, c, w, h, bounding_rect, background_color=None):

        painter.setBrush(QBrush(QColor(c.red(), c.green(), c.blue(), 150)))
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, 8, 8)


class FlowTheme_ColorfulLight(FlowTheme_Colorful):
    name = 'colorful light'
    type_ = 'light'

    header_padding = (12, 0, 2, 2)

    exec_conn_color = QColor('#1f1f1f')

    data_conn_color = QColor('#1f1f1f')

    flow_background_brush = QBrush(QColor('#ffffff'))
    flow_background_grid = ('points', QColor('#dddddd'), 2, 20, 20)

    node_title_color = QColor('#1f1f1f')
    port_pin_pen_color = QColor('#1f1f1f')

    node_item_shadow_color = QColor('#cccccc')

    def paint_PI_label(self, node_gui, painter, option, type_, connected, label_str, node_color, bounding_rect):
        c = None
        if not connected:
            c = QColor('#1f1f1f')
        else:
            if type_ == 'exec':
                c = QColor('#cccccc')
            else:
                c = node_color

        self.paint_PI_label_default(painter, label_str, c, QFont("Segoe UI", 10), bounding_rect)

    def draw_NI_normal(self, node_gui, selected: bool, hovered: bool, painter, c, w, h, bounding_rect: QRectF, title_rect):

        background_color = c.lighter()
        background_color.setAlpha(150)

        header_color = QColor(c.red(), c.green(), c.blue(), 130).darker()

        rel_header_height = self.get_header_rect(w, h, title_rect).height() / h
        gradient = QLinearGradient(bounding_rect.topLeft(), bounding_rect.bottomLeft())
        gradient.setColorAt(0, header_color)
        gradient.setColorAt(rel_header_height, header_color)
        gradient.setColorAt(rel_header_height + 0.0001, background_color)
        gradient.setColorAt(1, background_color)

        painter.setBrush(gradient)
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, 7, 7)

    def draw_NI_small(self, node_gui, selected: bool, hovered: bool,
                      painter, c, w, h, bounding_rect, background_color=None):

        painter.setBrush(QBrush(QColor(c.red(), c.green(), c.blue(), 150)))
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, 8, 8)


class FlowTheme_Industrial(FlowTheme):
    name = 'Industrial'

    header_padding = (8, 0, 8, 0)  # (8, 6, 10, 2)

    exec_conn_color = QColor(255, 255, 255)
    exec_conn_width = 2
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor(255, 194, 45)
    data_conn_width = 2
    data_conn_pen_style = Qt.DashLine

    flow_background_brush = QBrush(QColor(19, 19, 19))
    flow_background_grid = ('points', QColor(80, 80, 80), 2, 30, 30)

    node_color = QColor(10, 10, 10, 250)
    node_item_shadow_color = QColor(0, 0, 0)

    def paint_NI_title_label(self, node_gui, selected, hovering, painter, option, node_style, node_title, node_color,
                             node_item_bounding_rect):
        if node_style == 'normal':
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=QColor(200, 200, 200),
                pen_w=1,
                font=QFont('Segoe UI', 11, QFont.Normal if not (hovering or selected) else QFont.Bold),
                node_item_bounding_rect=node_item_bounding_rect
            )
        else:
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=node_color,
                pen_w=2,
                font=QFont('Segoe UI', 15, QFont.Bold),
                node_item_bounding_rect=node_item_bounding_rect
            )

    def paint_PI_label(self, node_gui, painter, option, type_, connected, label_str, node_color, bounding_rect):
        c = QColor('#FFFFFF')
        self.paint_PI_label_default(painter, label_str, c, QFont("Segoe UI", 8, QFont.Normal), bounding_rect)

    def paint_PI(self, node_gui, painter, option, node_color, type_, connected, rect):

        color = QColor('#FFFFFF') if type_ == 'exec' else QColor(node_color)

        # add = 2
        # padd = padding + add
        outer_ellipse_rect = rect.marginsRemoved(QMarginsF(2, 2, 2, 2))  # QRectF(padd, padd, w - 2*padd, h - 2*padd)
        # add = 4
        # padd = padding + add
        inner_ellipse_rect = rect.marginsRemoved(QMarginsF(4, 4, 4, 4))  # QRectF(padd, padd, w - 2*padd, h - 2*padd)

        if type_ == 'exec':

            if connected:
                brush = QBrush(QColor(255, 255, 255, 200))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)

            rect_ = rect.marginsRemoved(QMarginsF(2, 2, 2, 2))

            painter.setPen(QPen(QColor(255, 255, 255)))

            painter.drawPolygon(QPolygon([
                rect_.topLeft().toPoint(),
                QPoint(rect_.right(), rect_.center().toPoint().y()),
                rect_.bottomLeft().toPoint(),
            ]))

        elif type_ == 'data':

            pen = QPen(color)
            pen.setWidth(1)
            painter.setPen(pen)

            if connected:
                # draw inner ellipse
                brush = QBrush(QColor(color.red(), color.green(), color.blue(), 200))
                painter.setBrush(brush)
                painter.drawEllipse(inner_ellipse_rect)

            # draw outer ellipse
            painter.setBrush(Qt.NoBrush)
            painter.drawEllipse(outer_ellipse_rect)

    def draw_NI_normal(self, node_gui, selected: bool, hovered: bool,
                       painter, c, w, h, bounding_rect, title_rect):

        background_color = QColor(14, 14, 14)
        header_color = QColor(105, 105, 105, 150)

        rel_header_height = self.get_header_rect(w, h, title_rect).height() / h
        gradient = QLinearGradient(bounding_rect.topLeft(), bounding_rect.bottomLeft())
        gradient.setColorAt(0, header_color)
        gradient.setColorAt(rel_header_height, header_color)
        gradient.setColorAt(rel_header_height + 0.0001, background_color)
        gradient.setColorAt(1, background_color)

        painter.setBrush(gradient)
        painter.setPen(Qt.NoPen)  # QPen(c.darker())
        painter.drawRoundedRect(bounding_rect, 2, 2)

    def draw_NI_small(self, node_gui, selected: bool, hovered: bool,
                      painter, c, w, h, bounding_rect, background_color=None):

        background_color = QColor(217, 217, 217, 50)
        c_s = 2
        painter.setBrush(self.interpolate_color(c, background_color, 0.97))
        pen = QPen(QColor(130, 130, 130))
        painter.setPen(pen)
        painter.drawRoundedRect(bounding_rect, c_s, c_s)


class FlowTheme_Fusion(FlowTheme):
    name = 'Fusion'
    type_ = 'light'

    node_selection_stylesheet = '''
    NodeSelectionWidget {
        background-color: white;
    }
    NodeWidget {
        background-color: white;
    }
        '''

    exec_conn_color = QColor('#1f1f1f')
    exec_conn_width = 2

    data_conn_color = QColor('#1f1f1f')
    data_conn_width = 2

    flow_background_brush = QBrush(QColor('#ffffff'))

    node_normal_bg_col = QColor('#ebeced')
    node_small_bg_col = QColor('#cccdcf')
    node_title_color = QColor('#1f1f1f')
    port_pin_pen_color = QColor('#1f1f1f')

    node_item_shadow_color = QColor('#cccccc')


    def paint_NI_title_label(self, node_gui, selected, hovering, painter, option, node_style, node_title, node_color,
                             node_item_bounding_rect):

        painter.setPen(QPen(self.node_title_color))

        if node_style == 'normal':
            painter.setFont(QFont('Segoe UI', 10))
            align = Qt.AlignLeft | Qt.AlignVCenter
        else:
            painter.setFont(QFont('Segoe UI', 12))
            align = Qt.AlignCenter

        painter.drawText(node_item_bounding_rect, align, node_title)


    def paint_PI_label(self, node_gui, painter, option, type_, connected, label_str, node_color, bounding_rect):
        pen = QPen(QColor('#000000'))
        pen.setWidthF(1.2)
        painter.setPen(pen)

        self.paint_PI_label_default(painter, label_str, QColor(0, 0, 0), QFont("Segoe UI", 8), bounding_rect)


    def paint_PI(self, node_gui, painter, option, node_color, type_, connected, rect):

        painter.setBrush(QColor('#000000'))
        painter.setPen(Qt.NoPen)

        if type_ == 'data':
            painter.drawEllipse(rect.marginsRemoved(QMarginsF(3, 3, 3, 3)))
        else:
            draw_rect = rect.marginsRemoved(QMarginsF(3, 3, 3, 3))
            path = QPainterPath(draw_rect.topLeft())
            path.lineTo(QPointF(draw_rect.right(), draw_rect.center().y()))
            path.lineTo(draw_rect.bottomLeft())
            path.closeSubpath()

            painter.drawPath(path)


    def draw_NI_normal(self, node_gui, selected: bool, hovered: bool,
                       painter, c, w, h, bounding_rect: QRectF, title_rect):

        pen = QPen(c)
        col_top = self.node_normal_bg_col.lighter(105)
        col_bottom = self.node_normal_bg_col

        if not selected:
            pen.setWidthF(1)
        else:
            pen.setWidthF(2.5)
            col_bottom = QColor(255, 255, 255)

        header_height = self.get_header_rect(w, h, title_rect).height()
        header_fraction = header_height / bounding_rect.height()

        gradient = QLinearGradient(bounding_rect.topLeft(), bounding_rect.bottomLeft())
        gradient.setColorAt(0, col_top)
        gradient.setColorAt(header_fraction, self.interpolate_color(col_top, col_bottom, 0.7))
        gradient.setColorAt(1, col_bottom)

        painter.setBrush(QBrush(gradient))
        painter.setPen(pen)
        painter.drawRoundedRect(bounding_rect, 3, 3)


    def draw_NI_small(self, node_gui, selected: bool, hovered: bool,
                      painter, c, w, h, bounding_rect, background_color=None):

        painter.setBrush(QBrush(self.node_small_bg_col))
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, 4, 4)


flow_themes: List[FlowTheme] = [
    FlowTheme_Toy(),
    FlowTheme_DarkTron(),
    FlowTheme_Ghost(),
    FlowTheme_Blender(),
    FlowTheme_Simple(),
    FlowTheme_Ueli(),
    FlowTheme_PureDark(),
    FlowTheme_Colorful(),
    FlowTheme_PureLight(),
    FlowTheme_ColorfulLight(),
    FlowTheme_Industrial(),
    FlowTheme_Fusion(),
]



# list widgets
from .VariablesListWidget import VariablesListWidget as VarsList
from .FlowsListWidget import FlowsListWidget as FlowsList
from ..flows.node_list_widget.NodeListWidget import NodeListWidget

# logging
from .LogWidget import LogWidget



from typing import List, Dict

from qtpy.QtCore import QObject, Signal, Qt
from qtpy.QtWidgets import QWidget, QApplication

import ryvencore

from .flows.FlowView import FlowView
from .Design import Design
from .GUIBase import GUIBase


class SessionGUI(GUIBase, QObject):
    """
    ryvencore-qt's Session wrapper class, implementing the GUI.
    Any session with a GUI must be created through this class.
    Access the ryvencore session through the :code:`session`
    attribute, and the GUI from the ryvencore session through the
    :code:`gui` attribute. Once instantiated, you can simply use
    the :code:`session` directly to create, rename, delete flows,
    register nodes, etc.
    """

    flow_created = Signal(object)
    flow_deleted = Signal(object)
    flow_renamed = Signal(object, str)
    flow_view_created = Signal(object, object)

    def __init__(self, gui_parent: QWidget):
        GUIBase.__init__(self)
        QObject.__init__(self)

        self.core_session = ryvencore.Session(gui=True, load_addons=True)
        setattr(self.core_session, 'gui', self)

        self.gui_parent = gui_parent

        # flow views
        self.flow_views: Dict[ryvencore.Flow, FlowView] = {}

        # register complete_data function
        ryvencore.set_complete_data_func(self.get_complete_data_function(self))

        # load design
        app = QApplication.instance()
        app.setAttribute(Qt.AA_UseHighDpiPixmaps)
        Design.register_fonts()
        self.design = Design()

        # connect to session
        self.core_session.flow_created.sub(self._flow_created)
        self.core_session.flow_deleted.sub(self._flow_deleted)
        self.core_session.flow_renamed.sub(self._flow_renamed)

    def _flow_created(self, flow: ryvencore.Flow):
        """
        Builds the flow view for a newly created flow, saves it in
        self.flow_views, and emits the flow_view_created signal.
        """
        self.flow_created.emit(flow)

        self.flow_views[flow] = FlowView(
            session_gui=self,
            flow=flow,
            parent=self.gui_parent,
        )
        self.flow_view_created.emit(flow, self.flow_views[flow])

        return flow

    def _flow_deleted(self, flow: ryvencore.Flow):
        """
        Removes the flow view for a deleted flow from self.flow_views.
        """
        self.flow_views.pop(flow)
        self.flow_deleted.emit(flow)

    def _flow_renamed(self, flow: ryvencore.Flow, new_name: str):
        """
        Renames the flow view for a renamed flow.
        """
        self.flow_renamed.emit(flow, new_name)


from qtpy.QtCore import Qt
from qtpy.QtWidgets import QWidget, QMessageBox, QVBoxLayout, QLineEdit, QHBoxLayout, QPushButton, QScrollArea

from .FlowsList_FlowWidget import FlowsList_FlowWidget


class FlowsListWidget(QWidget):
    """Convenience class for a QWidget to easily manage the flows of a session."""

    def __init__(self, session_gui):
        super().__init__()

        self.session_gui = session_gui
        self.list_widgets = []
        self.ignore_name_line_edit_signal = False  # because disabling causes firing twice otherwise

        self.setup_UI()

        self.session_gui.flow_view_created.connect(self.add_new_flow)
        self.session_gui.flow_deleted.connect(self.recreate_list)


    def setup_UI(self):
        main_layout = QVBoxLayout(self)
        main_layout.setAlignment(Qt.AlignTop)
        self.setLayout(main_layout)

        # list scroll area

        self.list_scroll_area = QScrollArea(self)
        self.list_scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.list_scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.list_scroll_area.setWidgetResizable(True)
        self.list_scroll_area.setContentsMargins(0, 0, 0, 0)

        self.scroll_area_widget = QWidget()
        self.scroll_area_widget.setContentsMargins(0, 0, 0, 0)
        self.list_scroll_area.setWidget(self.scroll_area_widget)

        self.list_layout = QVBoxLayout()
        self.list_layout.setContentsMargins(0, 0, 0, 0)
        self.list_layout.setAlignment(Qt.AlignTop)
        self.scroll_area_widget.setLayout(self.list_layout)

        self.layout().addWidget(self.list_scroll_area)

        # line edit

        self.new_flow_title_lineedit = QLineEdit()
        self.new_flow_title_lineedit.setPlaceholderText('new flow\'s title')
        self.new_flow_title_lineedit.returnPressed.connect(self.create_flow)

        main_layout.addWidget(self.new_flow_title_lineedit)


        self.recreate_list()


    def recreate_list(self):
        # remove flow widgets
        for i in reversed(range(self.list_layout.count())):
            self.list_layout.itemAt(i).widget().setParent(None)
        self.list_widgets.clear()

        # re-create flow widgets
        for s in self.session_gui.core_session.flows:
            new_widget = FlowsList_FlowWidget(self, self.session_gui, s)
            self.list_widgets.append(new_widget)

        # add flow widgets to layout
        for w in self.list_widgets:
            self.list_layout.addWidget(w)

    def create_flow(self):
        title = self.new_flow_title_lineedit.text()

        if self.session_gui.core_session.flow_title_valid(title):
            self.session_gui.core_session.create_flow(title=title)

    def add_new_flow(self, flow, flow_view):
        self.recreate_list()

    def del_flow(self, flow, flow_widget):
        msg_box = QMessageBox(QMessageBox.Warning, 'sure about deleting flow?',
                              'You are about to delete a flow. This cannot be undone, all content will be lost. '
                              'Do you want to continue?', QMessageBox.Cancel | QMessageBox.Yes, self)
        msg_box.setDefaultButton(QMessageBox.Cancel)
        ret = msg_box.exec_()
        if ret != QMessageBox.Yes:
            return

        self.list_widgets.remove(flow_widget)
        flow_widget.setParent(None)
        self.session_gui.core_session.delete_flow(flow)
        # self.recreate_list()



"""The base classes for node custom widgets for nodes."""
from typing import Dict
from ryvencore import Data

from ..FlowCommands import Delegate_Command


class NodeMainWidget:
    """Base class for the main widget of a node."""

    def __init__(self, params):
        self.node, self.node_item, self.node_gui = params

    def get_state(self) -> Dict:
        """
        *VIRTUAL*

        Return the state of the widget, in a (pickle) serializable format.
        """
        data: Dict = {}
        return data

    def set_state(self, data: Dict):
        """
        *VIRTUAL*

        Set the state of the widget, where data corresponds to the dict
        returned by get_state().
        """
        pass

    def update_node(self):
        self.node.update()

    def update_node_shape(self):
        self.node_item.update_shape()


class NodeInputWidget:
    """Base class for the input widget of a node."""

    def __init__(self, params):
        self.input, self.input_item, self.node, self.node_gui, self.position = \
            params

    def get_state(self) -> Dict:
        """
        *VIRTUAL*

        Return the state of the widget, in a (pickle) serializable format.
        """
        data: Dict = {}
        return data

    def set_state(self, data: Dict):
        """
        *VIRTUAL*

        Set the state of the widget, where data corresponds to the dict
        returned by get_state().
        """
        pass

    def val_update_event(self, val: Data):
        """
        *VIRTUAL*

        Called when the input's value is updated through a connection.
        This can be used to represent the value in the widget.
        The widget is disabled when the port is connected.
        """
        pass

    # API methods

    def update_node_input(self, val: Data, silent=False):
        """
        Update the input's value and update the node.
        """
        self.input.default = val
        if not silent:
            self.input.node.update(self.node.inputs.index(self.input))

    def update_node(self):
        self.node.update(self.node.inputs.index(self.input))

    def update_node_shape(self):
        self.node_gui.update_shape()


class NodeInspectorWidget:
    """Base class for the inspector widget of a node."""

    def __init__(self, params):
        self.node, self.node_gui = params

    def get_state(self) -> Dict:
        """
        *VIRTUAL*

        Return the state of the widget, in a (pickle) serializable format.
        """
        data: Dict = {}
        return data

    def set_state(self, data: Dict):
        """
        *VIRTUAL*

        Set the state of the widget, where data corresponds to the dict
        returned by get_state().
        """
        pass

    def load(self):
        """Called when the inspector is loaded into the inspector view in the editor."""
        pass

    def unload(self):
        """Called when the inspector is removed from the inspector view in the editor."""
        pass

    def push_undo(self, text: str, undo_fn, redo_fn):
        """Push an undo function to the undo stack of the flow."""
        self.node_gui.flow_view().push_undo(
            Delegate_Command(
                self.node_gui.flow_view(),
                text=text,
                on_undo=undo_fn,
                on_redo=redo_fn,
            )
        )



# prevent circular imports
from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .NodeGUI import NodeGUI
    from .NodeItem import NodeItem
    from ..FlowView import FlowView

from typing import Tuple, Optional

from qtpy.QtWidgets import QGraphicsGridLayout, QGraphicsWidget, QGraphicsLayoutItem
from qtpy.QtCore import Qt, QRectF, QPointF, QSizeF
from qtpy.QtGui import QFontMetricsF, QFont

from ryvencore import serialize, Data
from ryvencore.NodePort import NodeOutput, NodeInput, NodePort
from ryvencore.utils import deserialize

from ...GUIBase import GUIBase
from ...utils import get_longest_line, shorten
from ..FlowViewProxyWidget import FlowViewProxyWidget
from .WidgetBaseClasses import NodeInputWidget

# utils


def is_connected(port):
    if isinstance(port, NodeOutput):
        is_connected = len(port.node.flow.connected_inputs(port)) > 0
    else:
        is_connected = port.node.flow.connected_output(port) is not None
    return is_connected


def val(port):
    if isinstance(port, NodeOutput):
        return port.val.payload if isinstance(port.val, Data) else None
    else:
        conn_out = port.node.flow.connected_output(port)
        if conn_out:
            return conn_out.val.payload if conn_out.val is not None else None
        else:
            return None


def connections(port):
    if isinstance(port, NodeOutput):
        return [(port, i) for i in port.node.flow.connected_inputs(port)]
    else:
        conn_out = port.node.flow.connected_output(port)
        if conn_out:
            return [(port.node.flow.connected_output(port), port)]
        else:
            return []


# main classes


class PortItem(GUIBase, QGraphicsWidget):
    """The GUI representative for ports of nodes, also handling mouse events for connections."""

    def __init__(self, node_gui: NodeGUI, node_item: NodeItem, port: NodePort, flow_view: FlowView):
        GUIBase.__init__(self, representing_component=port)
        QGraphicsWidget.__init__(self)

        self.setGraphicsItem(self)

        self.node_gui: NodeGUI = node_gui
        self.node_item: NodeItem = node_item
        self.port: NodePort = port
        self.flow_view: FlowView = flow_view

        # self.port.has_been_connected.connect(self.port_connected)
        # self.port.has_been_disconnected.connect(self.port_disconnected)

        self.pin = PortItemPin(self.port, self, self.node_gui, self.node_item)

        self.label = PortItemLabel(self.port, self, self.node_gui, self.node_item)

        self._layout = QGraphicsGridLayout()
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._layout)

    # >>> interaction boilerplate >>>
    def boundingRect(self):
        return QRectF(QPointF(0, 0), self.geometry().size())

    def setGeometry(self, rect):
        self.prepareGeometryChange()
        QGraphicsLayoutItem.setGeometry(self, rect)
        self.setPos(rect.topLeft())
    # <<< interaction boilerplate <<<

    def setup_ui(self):
        pass

    def port_connected(self):
        pass

    def port_disconnected(self):
        pass


class InputPortItem(PortItem):
    def __init__(self, node_gui, node_item, port, input_widget: Optional[Tuple[type, str]] = None):
        self.port: NodeInput
        super().__init__(node_gui, node_item, port, node_gui.flow_view())

        self.proxy = None  # widget proxy
        self.widget: Optional[NodeInputWidget] = None  # widget
        if input_widget is not None:
            self.create_widget(input_widget[0], input_widget[1])

        self.update_widget_value = (
            self.widget is not None
        )  # modified by FlowView when performance mode changes

        # catch up to missed connections
        if self.port.node.flow.connected_output(self.port) is not None:
            self.port_connected()

        if (
            self.port.type_ == 'data'
            and self.port.load_data is not None
            and self.port.load_data['has widget']
        ):
            c_d = self.port.load_data['widget data']
            if c_d is not None:
                assert self.widget is not None
                self.widget.set_state(deserialize(c_d))
            else:
                # this is a little feature that lets us prevent loading of input widgets
                # which is occasionally useful, e.g. when changing an input widget class:
                # to prevent loading of the input widget, 'widget data' must be None
                pass

        self.setup_ui()

    def setup_ui(self):
        l = self._layout

        # l.setSpacing(0)
        l.addItem(self.pin, 0, 0)
        l.setAlignment(self.pin, Qt.AlignVCenter | Qt.AlignLeft)
        l.addItem(self.label, 0, 1)
        l.setAlignment(self.label, Qt.AlignVCenter | Qt.AlignLeft)
        if self.widget:
            if self.widget.position == 'below':
                l.addItem(self.proxy, 1, 0, 1, 2)
            elif self.widget.position == 'besides':
                l.addItem(self.proxy, 0, 2)
            else:
                print('Unknown input widget position:', self.widget.position)

            l.setAlignment(self.proxy, Qt.AlignCenter)

    def create_widget(self, widget_class, widget_pos):
        if widget_class is None:
            return

        if self.port.type_ != 'data':
            # TODO: how about input widgets for exec inputs?
            return

        params = (self.port, self, self.node_gui.node, self.node_gui, widget_pos)

        # custom input widget
        self.widget = widget_class(params)
        self.proxy = FlowViewProxyWidget(self.flow_view, parent=self.node_item)
        self.proxy.setWidget(self.widget)

    def port_connected(self):
        """Disables the widget"""
        if self.widget:
            self.widget.setEnabled(False)

        # https://github.com/leon-thomm/Ryven/pull/137#issuecomment-1433783052
        # if self.port.type_ == 'data':
        #     self.port.connections[0].activated.connect(self._port_val_updated)
        #
        # self._port_val_updated(self.port.val)

    def port_disconnected(self):
        """Enables the widget again"""
        if self.widget:
            self.widget.setEnabled(True)

    def complete_data(self, data: dict) -> dict:
        if self.port.type_ == 'data':
            if self.widget:
                data['has widget'] = True
                data['widget name'] = self.node_gui.input_widgets[self.port]['name']
                data['widget pos'] = self.node_gui.input_widgets[self.port]['pos']
                data['widget data'] = serialize(self.widget.get_state())
            else:
                data['has widget'] = False

        return data


class OutputPortItem(PortItem):
    def __init__(self, node_gui, node_item, port):
        super().__init__(node_gui, node_item, port, node_gui.flow_view())
        # super(OutputPortItem, self).__init__(parent_node_instance, PortObjPos.OUTPUT, type_, label_str)

        self.setup_ui()

    def setup_ui(self):
        l = self._layout

        # l.setSpacing(5)
        l.addItem(self.label, 0, 0)
        l.setAlignment(self.label, Qt.AlignVCenter | Qt.AlignRight)
        l.addItem(self.pin, 0, 1)

        l.setAlignment(self.pin, Qt.AlignVCenter | Qt.AlignRight)


# contents


class PortItemPin(QGraphicsWidget):
    def __init__(self, port, port_item, node_gui, node_item):
        super(PortItemPin, self).__init__(node_item)

        self.port = port
        self.port_item = port_item
        self.node_gui = node_gui
        self.node_item = node_item
        self.flow_view = self.node_item.flow_view

        self.setGraphicsItem(self)
        self.setAcceptHoverEvents(True)
        self.hovered = False
        self.setCursor(Qt.CrossCursor)
        self.tool_tip_pos = None

        self.padding = 2
        # self.painting_width = 17
        # self.painting_height = 17
        # self.width = self.painting_width+2*self.padding
        # self.height = self.painting_height+2*self.padding
        self.width = 17
        self.height = 17
        self.port_local_pos = None

    def boundingRect(self):
        return QRectF(QPointF(0, 0), self.geometry().size())

    def setGeometry(self, rect):
        self.prepareGeometryChange()
        QGraphicsLayoutItem.setGeometry(self, rect)
        self.setPos(rect.topLeft())

    def sizeHint(self, which, constraint=...):
        return QSizeF(self.width, self.height)

    def paint(self, painter, option, widget=None):
        self.node_item.session_design.flow_theme.paint_PI(
            node_gui=self.node_gui,
            painter=painter,
            option=option,
            node_color=self.node_gui.color,
            type_=self.port.type_,
            connected=is_connected(self.port),
            rect=QRectF(
                self.padding, self.padding, self.width_no_padding(), self.height_no_padding()
            ),
        )

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:  # DRAG NEW CONNECTION
            self.flow_view.mouse_event_taken = True
            self.flow_view._selected_pin = self
            self.flow_view._dragging_connection = True
            event.accept()  # don't pass the ev ent to anything below
        else:
            return QGraphicsWidget.mousePressEvent(self, event)

    def moveEvent(self, event):
        super().moveEvent(event)

        # update connections
        conn_items = self.flow_view.connection_items
        for c in self.port.connections:
            i = conn_items[c]

            # if the items are grouped (which means they move together), don't recompute
            if i.out.group() is None or i.out.group() != i.inp.group():  # not entirely sure if this is working
                i.recompute()

    def hoverEnterEvent(self, event):
        if self.port.type_ == 'data':  # and self.parent_port_instance.io_pos == PortPos.OUTPUT:
            self.setToolTip(shorten(str(val(self.port)), 1000, line_break=True))

        # highlight connections
        items = self.flow_view.connection_items
        for c in connections(self.port):
            items[c].set_highlighted(True)

        self.hovered = True

        QGraphicsWidget.hoverEnterEvent(self, event)

    def hoverLeaveEvent(self, event):
        # un-highlight connections
        items = self.flow_view.connection_items
        for c in connections(self.port):
            items[c].set_highlighted(False)

        self.hovered = False

        QGraphicsWidget.hoverLeaveEvent(self, event)

    def width_no_padding(self):
        """The width without the padding"""
        return self.width - 2 * self.padding

    def height_no_padding(self):
        """The height without the padding"""
        return self.height - 2 * self.padding

    def get_scene_center_pos(self) -> QPointF:
        if not self.node_item.collapsed:
            return QPointF(
                self.scenePos().x() + self.boundingRect().width() / 2,
                self.scenePos().y() + self.boundingRect().height() / 2,
            )
        else:
            # mypy seems buggy here, it things below methods return Any,
            # but they are annotated to return QPointF
            if isinstance(self.port_item, InputPortItem):
                return self.node_item.get_left_body_header_vertex_scene_pos()  # type: ignore
            else:
                return self.node_item.get_right_body_header_vertex_scene_pos()  # type: ignore


class PortItemLabel(QGraphicsWidget):
    def __init__(self, port, port_item, node_gui, node_item):
        super(PortItemLabel, self).__init__(node_item)
        self.setGraphicsItem(self)

        self.port = port
        self.port_item = port_item
        self.node_gui = node_gui
        self.node_item = node_item

        self.font = QFont("Source Code Pro", 10, QFont.Bold)
        font_metrics = QFontMetricsF(
            self.font
        )  # approximately! the designs can use different fonts
        self.width = font_metrics.width(get_longest_line(self.port.label_str))
        self.height = font_metrics.height() * (self.port.label_str.count('\n') + 1)
        self.port_local_pos = None

    def boundingRect(self):
        return QRectF(QPointF(0, 0), self.geometry().size())

    def setGeometry(self, rect):
        self.prepareGeometryChange()
        QGraphicsLayoutItem.setGeometry(self, rect)
        self.setPos(rect.topLeft())

    def sizeHint(self, which, constraint=...):
        return QSizeF(self.width, self.height)

    def paint(self, painter, option, widget=None):
        self.node_item.session_design.flow_theme.paint_PI_label(
            self.node_gui,
            painter,
            option,
            self.port.type_,
            is_connected(self.port),
            self.port.label_str,
            self.node_gui.color,
            self.boundingRect(),
        )



from qtpy.QtGui import QFontMetrics, QFont
from qtpy.QtWidgets import QSpinBox, QLineEdit, QCheckBox, QComboBox

from .WidgetBaseClasses import NodeInputWidget


class DType_IW_Base(NodeInputWidget):

    def __init__(self, params):
        super().__init__(params)

        self.dtype = self.input.dtype
        self.block = False

    def change_val(self, val):
        if not self.block:
            self.dtype.val = val
            self.update_node_input(val)

    def __str__(self):
        return self.__class__.__name__


class Data_IW(DType_IW_Base, QLineEdit):  # virtual

    base_width: int  # specified by subclasses

    def __init__(self, params):
        DType_IW_Base.__init__(self, params)
        QLineEdit.__init__(self)

        # dtype = self.input.dtype
        # self.last_val = None

        self.setFont(QFont('source code pro', 10))
        self.val_update_event(self.dtype.val)
        if self.dtype.size == 's':
            self.base_width = 30
        elif self.dtype.size == 'm':
            self.base_width = 70
        elif self.dtype.size == 'l':
            self.base_width = 150
        self.max_width = self.base_width * 3
        self.setFixedWidth(self.base_width)
        self.fm = QFontMetrics(self.font())

        self.setToolTip(self.dtype.doc)
        self.textChanged.connect(self.text_changed)
        self.editingFinished.connect(self.editing_finished)

    def text_changed(self, new_text):
        """manages resizing of widget to content"""

        text_width = self.fm.width(new_text)
        new_width = text_width+15  # add some buffer
        if new_width < self.max_width:
            self.setFixedWidth(new_width if new_width > self.base_width else self.base_width)
        else:
            self.setFixedWidth(self.max_width)
        self.node.update_shape()

    def editing_finished(self):
        """updates the input"""
        # if v != self.last_val:
        # self.last_val = v
        self.change_val(self.get_val())

    def get_val(self):
        try:
            return eval(self.text())
        except Exception as e:
            return self.text()

    def val_update_event(self, val):
        """triggered when input is connected and received new data;
        displays the data in the widget (without updating)"""

        self.block = True
        try:
            self.setText(str(val))
        except Exception as e:
            pass
        finally:
            self.block = False

    def get_state(self) -> dict:
        return {'text': self.text()}

    def set_state(self, data: dict):
        # just show value, DO NOT UPDATE
        self.val_update_event(data['text'])


# custom sized classes for qss access:

class Data_IW_S(Data_IW):
    base_width = 30


class Data_IW_M(Data_IW):
    base_width = 70


class Data_IW_L(Data_IW):
    base_width = 150


# -----------------------------------


class String_IW(DType_IW_Base, QLineEdit):  # virtual

    width_: int   # specified by subclasses

    def __init__(self, params):
        DType_IW_Base.__init__(self, params)
        QLineEdit.__init__(self)

        # dtype = self.input.dtype
        # self.last_val = None

        self.setFont(QFont('source code pro', 10))
        self.setText(self.dtype.val)
        self.setFixedWidth(self.width_)
        self.setToolTip(self.dtype.doc)

        self.editingFinished.connect(self.editing_finished)

    def editing_finished(self):
        """updates the input"""
        self.change_val(self.get_val())
        # v = self.get_val()
        # if v != self.last_val:
        #     self.update_node_input(v)
        #     self.last_val = v

    def get_val(self):
        return self.text()

    def val_update_event(self, val):
        """triggered when input is connected and received new data;
        displays the data in the widget (without updating)"""
        self.block = True
        self.setText(str(val))
        self.block = False

    def get_state(self) -> dict:
        return {'text': self.text()}

    def set_state(self, data: dict):
        # just show value, DO NOT UPDATE
        self.val_update_event(data['text'])


# custom sized classes for qss access:

class String_IW_S(String_IW):
    width_ = 30


class String_IW_M(String_IW):
    width_ = 70


class String_IW_L(String_IW):
    width_ = 150


# -----------------------------------


class Integer_IW(DType_IW_Base, QSpinBox):
    def __init__(self, params):
        DType_IW_Base.__init__(self, params)
        QSpinBox.__init__(self)

        if self.dtype.bounds:
            self.setRange(self.dtype.bounds[0], self.dtype.bounds[1])
        self.setValue(self.dtype.val)
        self.setToolTip(self.dtype.doc)

        self.valueChanged.connect(self.widget_val_changed)

    def widget_val_changed(self, val):
        self.change_val(val)

    def get_val(self):
        return self.value()

    def val_update_event(self, val):
        """triggered when input is connected and received new data;
        displays the data in the widget (without updating)"""
        self.block = True
        try:
            self.setValue(val)
        except Exception as e:
            pass
        finally:
            self.block = False

    def get_state(self) -> dict:
        return {'val': self.value()}

    def set_state(self, data: dict):
        # just show value, DO NOT UPDATE
        self.val_update_event(data['val'])


class Float_IW(DType_IW_Base, QLineEdit):
    def __init__(self, params):
        DType_IW_Base.__init__(self, params)
        QLineEdit.__init__(self)

        self.setFont(QFont('source code pro', 10))
        fm = QFontMetrics(self.font())
        self.setMaximumWidth(fm.width(' ')*self.dtype.decimals+1)
        self.setText(str(self.dtype.val))
        self.setToolTip(self.dtype.doc)

        self.textChanged.connect(self.widget_text_changed)

    def widget_text_changed(self):
        self.change_val(self.get_val())

    def get_val(self):
        return float(self.text())

    def val_update_event(self, val):
        """triggered when input is connected and received new data;
        displays the data in the widget (without updating)"""
        self.block = True
        try:
            self.setText(str(val))
        except Exception as e:
            pass
        finally:
            self.block = False

    def get_state(self) -> dict:
        return {'text': self.text()}

    def set_state(self, data: dict):
        # just show value, DO NOT UPDATE
        self.val_update_event(data['text'])


class Boolean_IW(DType_IW_Base, QCheckBox):
    def __init__(self, params):
        DType_IW_Base.__init__(self, params)
        QCheckBox.__init__(self)

        self.setChecked(self.dtype.val)

        self.setToolTip(self.dtype.doc)

        self.stateChanged.connect(self.state_changed)

    def state_changed(self, state):
        self.change_val(self.get_val())

    def get_val(self):
        return self.isChecked()

    def val_update_event(self, val):
        """triggered when input is connected and received new data;
        displays the data in the widget (without updating)"""
        self.block = True
        try:
            self.setChecked(bool(val))
        except Exception as e:
            pass
        finally:
            self.block = False

    def get_state(self) -> dict:
        return {'checked': self.isChecked()}

    def set_state(self, data: dict):
        # just show value, DO NOT UPDATE
        self.val_update_event(data['checked'])


class Choice_IW(DType_IW_Base, QComboBox):
    def __init__(self, params):
        DType_IW_Base.__init__(self, params)
        QComboBox.__init__(self)

        self.addItems(self.dtype.items)
        self.setCurrentText(self.dtype.val)
        self.setToolTip(self.dtype.doc)

        self.currentTextChanged.connect(self.widget_text_changed)

    def widget_text_changed(self):
        self.change_val(self.get_val())

    def get_val(self):
        return self.currentText()

    def val_update_event(self, val):
        """triggered when input is connected and received new data;
        displays the data in the widget (without updating)"""
        self.block = True
        try:
            self.setCurrentText(val)
        except Exception as e:
            pass
        finally:
            self.block = False

    def get_state(self) -> dict:
        return {
            'items': [self.itemText(i) for i in range(self.count())],
            'active': self.currentText(),
        }

    def set_state(self, data: dict):
        # just show value, DO NOT UPDATE
        self.block = True
        self.clear()
        self.addItems(data['items'])
        self.setCurrentText(data['active'])
        self.block = False



# set package path (for resources etc.)
import os
from .src.GlobalAttributes import Location
Location.PACKAGE_PATH = os.path.normpath(os.path.dirname(__file__))

os.environ['RC_MODE'] = 'gui'  # set ryvencore gui mode
os.environ['QT_ENABLE_HIGHDPI_SCALING'] = '1'

# expose ryvencore
import ryvencore

from .src.SessionGUI import SessionGUI
from .src.flows.nodes.NodeGUI import NodeGUI

# customer base classes
from ryvencore import Node
from .src.flows.nodes.WidgetBaseClasses import NodeMainWidget, NodeInputWidget, NodeInspectorWidget

# gui classes
from .src.widgets import *
from .src.flows.FlowTheme import flow_themes



import traceback

from qtpy.QtWidgets import QGraphicsPixmapItem
from qtpy.QtGui import QPixmap, QImage

from ...GUIBase import GUIBase
from ...utils import get_resource


class NodeErrorIndicator(GUIBase, QGraphicsPixmapItem):

    def __init__(self, node_item):
        GUIBase.__init__(self)
        QGraphicsPixmapItem.__init__(self, parent=node_item)

        self.node = node_item
        self.pix = QPixmap(str(get_resource('pics/warning.png')))
        self.setPixmap(self.pix)
        self.setScale(0.1)
        self.setOffset(-self.boundingRect().width()/2, -self.boundingRect().width()/2)

    def set_error(self, e):
        error_msg = ''.join([
            f'<p>{line}</p>'
            for line in traceback.format_exc().splitlines()
        ])

        self.setToolTip(
            f'<html><head/><body>'
            f'{error_msg}'
            f'</body></html>'
        )



from typing import Union, Type, List, Optional, Tuple, TypeVar

from qtpy.QtWidgets import (
    QWidget, 
    QVBoxLayout, 
    QLabel, 
    QTextEdit,
    QSplitter,
)

from qtpy.QtCore import Qt
from ryvencore import Node

from .WidgetBaseClasses import NodeInspectorWidget


class InspectorView(QWidget):
    """
    A widget that can display the inspector of the currently selected node.
    """

    def __init__(self, flow_view, parent: Optional[QWidget] = None):
        super().__init__(parent=parent)
        self.node: Optional[Node] = None
        self.inspector_widget: Optional[NodeInspectorWidget] = None
        self.flow_view = flow_view

        self.setup_ui()
        self.flow_view.nodes_selection_changed.connect(self.set_selected_nodes)

    def setup_ui(self):
        self.setLayout(QVBoxLayout())

    def set_selected_nodes(self, nodes: List[Node]):
        if len(nodes) == 0:
            self.set_node(None)
        else:
            self.set_node(nodes[-1])

    def set_node(self, node: Optional[Node]):
        """Sets a node for inspection, if it exists. Otherwise clears the inspector"""

        if self.node == node:
            return

        if self.inspector_widget is not None:
            assert isinstance(self.inspector_widget, QWidget)
            self.inspector_widget.setVisible(False)
            self.inspector_widget.setParent(None)  # type: ignore
            self.inspector_widget.unload()
            
        self.node = None
        self.inspector_widget = None

        if node is not None:
            self.node = node
            assert hasattr(self.node, 'gui')
            self.inspector_widget = self.node.gui.inspector_widget
            assert isinstance(self.inspector_widget, QWidget)
            self.layout().addWidget(self.inspector_widget)
            self.inspector_widget.load()
            self.inspector_widget.setVisible(True)


class NodeInspectorDefaultWidget(NodeInspectorWidget, QWidget):
    """
    Default node inspector widget implementation.
    Can also be extended by embedding a custom widget.
    """
    
    @staticmethod
    def _big_bold_text(txt: str):
        return f'<b><bold>{txt}</bold></b>'
    
    def __init__(self, params, child: Optional[NodeInspectorWidget] = None):
        QWidget.__init__(self)
        NodeInspectorWidget.__init__(self, params)

        self.child = child
        self.setLayout(QVBoxLayout())

        self.title_label: QLabel = QLabel()
        self.title_label.setText(
            f'<h2>{self.node.title}</h2> '
            f'<h4>id: {self.node.global_id}, pyid: {id(self.node)}</h4>'
        )
        # title
        self.layout().addWidget(self.title_label)
        
        # content splitter
        self.content_splitter = QSplitter()
        self.content_splitter.setOrientation(Qt.Orientation.Vertical)
        self.layout().addWidget(self.content_splitter)
        
        if child is not None:
            assert isinstance(child, QWidget)
            self.content_splitter.addWidget(child)

        desc = self.node.__doc__ if self.node.__doc__ and self.node.__doc__ != "" else "No description given"
        bbt = NodeInspectorDefaultWidget._big_bold_text
        
        self.description_area: QTextEdit = QTextEdit()
        self.description_area.setReadOnly(True)
        self.description_area.setText(f"""
<html>
    <body>
        {bbt('Title:')} {self.node.title}<br>
        {bbt('Version:')} {self.node.version}<br><br>
        {bbt('Description:')}<br><br>
        {desc}
    </body>
</html>
        """)
    
        self.content_splitter.addWidget(self.description_area)
    
    def load(self):
        super().load()
        if self.child:
            self.child.load()
    
    def unload(self):
        if self.child:
            self.child.unload()
        super().unload()



from queue import Queue
from typing import List, Dict, Tuple, Optional, Union, Type

from qtpy.QtCore import QObject, Signal

from .WidgetBaseClasses import NodeMainWidget, NodeInputWidget, NodeInspectorWidget
from .NodeInspector import NodeInspectorDefaultWidget


class NodeGUI(QObject):
    """
    Interface class between nodes and their GUI representation.
    """

    # customizable gui attributes
    description_html: Optional[str] = None
    main_widget_class: Optional[Type[NodeMainWidget]] = None
    main_widget_pos: str = 'below ports'
    input_widget_classes: Dict[str, Type[NodeInputWidget]] = {}
    inspector_widget_class: Type[NodeInspectorWidget] = NodeInspectorDefaultWidget
    wrap_inspector_in_default: bool = False
    init_input_widgets: dict = {}
    style: str = 'normal'
    color: str = '#c69a15'
    display_title: Optional[str] = None
    icon: Optional[str] = None

    # qt signals
    updating = Signal()
    update_error = Signal(object)
    input_added = Signal(int, object)
    output_added = Signal(int, object)
    input_removed = Signal(int, object)
    output_removed = Signal(int, object)
    update_shape_triggered = Signal()
    hide_unconnected_ports_triggered = Signal()
    show_unconnected_ports_triggered = Signal()

    def __init__(self, params):
        QObject.__init__(self)

        node, session_gui = params
        self.node = node
        self.item = None   # set by the node item directly after this __init__ call
        self.session_gui = session_gui
        setattr(node, 'gui', self)

        self.actions = self._init_default_actions()

        if self.display_title is None:
            self.display_title = self.node.title

        self.input_widgets = {}     # {input: widget name}
        for i, widget_data in self.init_input_widgets.items():
            self.input_widgets[self.node.inputs[i]] = widget_data
        # using attach_input_widgets() one can buffer input widget
        # names for inputs that are about to get created
        self._next_input_widgets = Queue()

        self.error_during_update = False

        # turn ryvencore signals into Qt signals
        self.node.updating.sub(self._on_updating)
        self.node.update_error.sub(self._on_update_error)
        self.node.input_added.sub(self._on_new_input_added)
        self.node.output_added.sub(self._on_new_output_added)
        self.node.input_removed.sub(self._on_input_removed)
        self.node.output_removed.sub(self._on_output_removed)

        # create the inspector widget
        inspector_params = (self.node, self)
        if self.wrap_inspector_in_default:
            self.inspector_widget = NodeInspectorDefaultWidget(
                child=self.inspector_widget_class((self.node, self)),
                params=inspector_params,
            )
        else:
            self.inspector_widget = self.inspector_widget_class(inspector_params)

    def initialized(self):
        """
        *VIRTUAL*

        Called after the node GUI has been fully initialized.
        The Node has been created already (including all ports) and loaded.
        No connections have been made to ports of the node yet.
        """
        pass

    """
    slots
    """

    # TODO: displaying update errors is currently prevented by the
    #   lack of an appropriate updated event in ryvencore.
    #   Update: there is an updating event now.

    # def on_updated(self, inp):
    #     if self.error_during_update:
    #         # an error should prevent an update event, so if we
    #         # are here, the update was successful
    #         self.self.error_during_update = False
    #         self.item.remove_error_message()
    #     self.updated.emit()
    #
    def _on_update_error(self, e):
    #     self.item.display_error(e)
    #     self.error_during_update = True
        self.update_error.emit(e)

    def _on_updating(self, inp: int):
        # update input widget
        if inp != -1 and self.item.inputs[inp].widget is not None:
            o = self.node.flow.connected_output(self.node.inputs[inp])
            if o is not None:
                self.item.inputs[inp].widget.val_update_event(o.val)

        self.updating.emit()

    def _on_new_input_added(self, _, index, inp):
        if not self._next_input_widgets.empty():
            self.input_widgets[inp] = self._next_input_widgets.get()
        self.input_added.emit(index, inp)

    def _on_new_output_added(self, _, index, out):
        self.output_added.emit(index, out)

    def _on_input_removed(self, _, index, inp):
        self.input_removed.emit(index, inp)

    def _on_output_removed(self, _, index, out):
        self.output_removed.emit(index, out)

    """
    actions
    
    TODO: move actions to ryvencore?
    """

    def _init_default_actions(self):
        """
        Returns the default actions every node should have
        """
        return {
            'update shape': {'method': self.update_shape},
            'hide unconnected ports': {'method': self.hide_unconnected_ports},
            'change title': {'method': self.change_title},
        }

    def _deserialize_actions(self, actions_data):
        """
        Recursively reconstructs the actions dict from the serialized version
        """

        def _transform(actions_data: dict):
            """
            Mutates the actions_data argument by replacing the method names
            with the actual methods. Doesn't modify the original dict.
            """
            new_actions = {}
            for key, value in actions_data.items():
                if key == 'method':
                    try:
                        value = getattr(self, value)
                    except AttributeError:
                        print(f'Warning: action method "{value}" not found in node "{self.node.title}", skipping.')
                elif isinstance(value, dict):
                    value = _transform(value)
                new_actions[key] = value
            return new_actions

        return _transform(actions_data)

    def _serialize_actions(self, actions):
        """
        Recursively transforms the actions dict into a JSON-compatible dict
        by replacing methods with their name. Doesn't modify the original dict.
        """

        def _transform(actions: dict):
            new_actions = {}
            for key, value in actions.items():
                if key == 'method':
                    new_actions[key] = value.__name__
                elif isinstance(value, dict):
                    new_actions[key] = _transform(value)
                else:
                    new_actions[key] = value
            return new_actions

        return _transform(actions)

    """
    serialization
    """

    def data(self):
        return {
            'actions': self._serialize_actions(self.actions),
            'display title': self.display_title,
            'inspector widget': self.inspector_widget.get_state(),
        }

    def load(self, data):
        if 'actions' in data:   # otherwise keep default
            self.actions = self._deserialize_actions(data['actions'])
        if 'display title' in data:
            self.display_title = data['display title']
        if 'special actions' in data:   # backward compatibility
            self.actions = self._deserialize_actions(data['special actions'])
        if 'inspector widget' in data:
            self.inspector_widget.set_state(data['inspector widget'])

    """
    GUI access methods
    """

    def set_display_title(self, t: str):
        self.display_title = t
        self.update_shape()

    def flow_view(self):
        return self.item.flow_view

    def main_widget(self):
        """Returns the main_widget object, or None if the item doesn't exist (yet)"""

        return self.item.main_widget

    def attach_input_widgets(self, widget_names: List[str]):
        """Attaches the input widget to the next created input."""

        for w in widget_names:
            self._next_input_widgets.queue(w)

    def input_widget(self, index: int):
        """Returns a reference to the widget of the corresponding input"""

        return self.item.inputs[index].widget

    def session_stylesheet(self):
        return self.session_gui.design.global_stylesheet

    def update_shape(self):
        """Causes recompilation of the whole shape of the GUI item."""

        self.update_shape_triggered.emit()

    def hide_unconnected_ports(self):
        """Hides all ports that are not connected to anything."""

        del self.actions['hide unconnected ports']
        self.actions['show unconnected ports'] = {'method': self.show_unconnected_ports}
        self.hide_unconnected_ports_triggered.emit()

    def show_unconnected_ports(self):
        """Shows all ports that are not connected to anything."""

        del self.actions['show unconnected ports']
        self.actions['hide unconnected ports'] = {'method': self.hide_unconnected_ports}
        self.show_unconnected_ports_triggered.emit()

    def change_title(self):
        from qtpy.QtWidgets import QDialog, QVBoxLayout, QLineEdit

        class ChangeTitleDialog(QDialog):
            def __init__(self, title):
                super().__init__()
                self.new_title = None
                self.setLayout(QVBoxLayout())
                self.line_edit = QLineEdit(title)
                self.layout().addWidget(self.line_edit)
                self.line_edit.returnPressed.connect(self.return_pressed)

            def return_pressed(self):
                self.new_title = self.line_edit.text()
                self.accept()

        d = ChangeTitleDialog(self.display_title)
        d.exec_()
        if d.new_title:
            self.set_display_title(d.new_title)
